import asyncio
import time

import pytest

from nanasqlite.utils import ExpirationMode, ExpiringDict


def test_expiring_dict_lazy():
    """LAZYモード（アクセス時のみチェック）のテスト"""
    d = ExpiringDict(expiration_time=0.5, mode=ExpirationMode.LAZY)
    d["key1"] = "val1"

    assert "key1" in d
    assert d["key1"] == "val1"

    time.sleep(0.6)

    # key1 は期限切れだが、__contains__ や __getitem__ 前まで内部的には残っている可能性がある
    # ただし ExpiringDict は __contains__ や __getitem__ 内でチェックする
    assert "key1" not in d
    with pytest.raises(KeyError):
        _ = d["key1"]

def test_expiring_dict_scheduler():
    """SCHEDULERモード（バックグラウンドワーカー）のテスト"""
    expired_keys = []
    def on_expire(k, v):
        expired_keys.append(k)

    d = ExpiringDict(expiration_time=0.3, mode=ExpirationMode.SCHEDULER, on_expire=on_expire)
    d["a"] = 1
    d["b"] = 2

    assert len(d) == 2

    # 0.3秒 + 余裕を持って待機
    time.sleep(0.8)

    # ワーカーによって削除されているはず
    assert "a" not in d
    assert "b" not in d
    assert len(d) == 0
    assert "a" in expired_keys
    assert "b" in expired_keys

    d.clear()

def test_expiring_dict_timer():
    """TIMERモード（個別タイマー）のテスト"""
    # 同期的な環境でも threading.Timer を使う
    expired_keys = []
    def on_expire(k, v):
        expired_keys.append(k)

    d = ExpiringDict(expiration_time=0.2, mode=ExpirationMode.TIMER, on_expire=on_expire)
    d["x"] = 10

    assert "x" in d
    time.sleep(0.5)

    assert "x" not in d
    assert "x" in expired_keys
    d.clear()

@pytest.mark.asyncio
async def test_expiring_dict_async_timer():
    """TIMERモードでの asyncio.call_later テスト"""
    expired_keys = []
    def on_expire(k, v):
        expired_keys.append(k)

    # loop がある環境では asyncio.call_later が優先されるはず
    d = ExpiringDict(expiration_time=0.1, mode=ExpirationMode.TIMER, on_expire=on_expire)
    d["y"] = 20

    assert "y" in d
    await asyncio.sleep(0.3)

    assert "y" not in d
    assert "y" in expired_keys
    d.clear()

def test_expiring_dict_update_ttl():
    """値を更新したときに有効期限がリセットされるかテスト"""
    d = ExpiringDict(expiration_time=0.5, mode=ExpirationMode.SCHEDULER)
    d["key"] = "v1"

    time.sleep(0.3)
    d["key"] = "v2" # リセット

    time.sleep(0.3)
    # 最初にセットしてから 0.6秒経っているが、途中で更新したのでまだ残っているはず
    assert "key" in d
    assert d["key"] == "v2"

    time.sleep(0.3)
    assert "key" not in d
    d.clear()

def test_expiring_dict_clear():
    """clear() でタイマーやワーカーが正しく処理されるかテスト"""
    d = ExpiringDict(expiration_time=1.0, mode=ExpirationMode.SCHEDULER)
    d["a"] = 1
    d.clear()
    assert len(d) == 0
    # 内部的に _scheduler_running が False になる
    assert not d._scheduler_running

def test_expiring_dict_pop():
    """pop() で削除された場合に期限切れコールバックが呼ばれないこと"""
    expired = []
    d = ExpiringDict(expiration_time=0.1, mode=ExpirationMode.SCHEDULER, on_expire=lambda k,v: expired.append(k))
    d["a"] = 1
    val = d.pop("a")
    assert val == 1
    time.sleep(0.2)
    assert len(expired) == 0
    d.clear()



"""TDD cycle 7 — Fix mypy rejection of types.EllipsisType.

Issue:
  mypy (python_version = "3.9") does not know types.EllipsisType, which was
  added in Python 3.10.  Running `mypy src/nanasqlite/core.py` produces:

      error: Name "types.EllipsisType" is not defined  [name-defined]

  Fix: update [tool.mypy] python_version from "3.9" to "3.10" in pyproject.toml
  (Python 3.9 reached EOL October 2025; the runtime is already 3.12).

TDD: all tests written BEFORE the fix is applied.
"""

import importlib.util
import shutil
import subprocess
import sys
from pathlib import Path

REPO_ROOT = Path(__file__).parent.parent
PYPROJECT = REPO_ROOT / "pyproject.toml"


def _run_mypy(*src_paths: str) -> subprocess.CompletedProcess:
    return subprocess.run(
        [sys.executable, "-m", "mypy", *src_paths],
        cwd=str(REPO_ROOT),
        capture_output=True,
        text=True,
    )


def _skip_if_mypy_not_installed() -> None:
    """Skip the current test if mypy is not available in the environment."""
    import pytest

    if shutil.which("mypy") is None and not importlib.util.find_spec("mypy"):
        pytest.skip("mypy not installed")


class TestMypyPassesWithEllipsisType:
    """mypy must not report name-defined errors for types.EllipsisType."""

    def test_mypy_installed(self):
        """Mypy must be importable (installed in the test environment)."""
        _skip_if_mypy_not_installed()
        assert importlib.util.find_spec("mypy") is not None, (
            "mypy is not installed — install it to run type-checking tests"
        )

    def test_mypy_no_name_defined_error_in_core(self):
        """mypy must not report 'Name types.EllipsisType is not defined' in core.py."""
        _skip_if_mypy_not_installed()

        result = _run_mypy("src/nanasqlite/core.py")
        name_defined_errors = [
            line
            for line in result.stdout.splitlines()
            if "EllipsisType" in line and "name-defined" in line
        ]
        assert name_defined_errors == [], (
            "mypy reports name-defined errors for types.EllipsisType in core.py:\n"
            + "\n".join(name_defined_errors)
            + "\nFix: update python_version to '3.10' in pyproject.toml [tool.mypy]"
        )

    def test_mypy_no_name_defined_error_in_async_core(self):
        """mypy must not report 'Name types.EllipsisType is not defined' in async_core.py."""
        _skip_if_mypy_not_installed()

        result = _run_mypy("src/nanasqlite/async_core.py")
        name_defined_errors = [
            line
            for line in result.stdout.splitlines()
            if "EllipsisType" in line and "name-defined" in line
        ]
        assert name_defined_errors == [], (
            "mypy reports name-defined errors for types.EllipsisType in async_core.py:\n"
            + "\n".join(name_defined_errors)
            + "\nFix: update python_version to '3.10' in pyproject.toml [tool.mypy]"
        )

    def test_mypy_python_version_in_config_is_at_least_310(self):
        """[tool.mypy] python_version must be >= '3.10' to support types.EllipsisType."""
        content = PYPROJECT.read_text()
        # Find the python_version setting under [tool.mypy]
        in_mypy_section = False
        for line in content.splitlines():
            if line.strip() == "[tool.mypy]":
                in_mypy_section = True
                continue
            if in_mypy_section and line.startswith("["):
                # Entered next section without finding python_version
                break
            if in_mypy_section and line.strip().startswith("python_version"):
                # e.g. python_version = "3.10"
                version_str = line.split("=", 1)[1].strip().strip('"').strip("'")
                major, minor = map(int, version_str.split(".")[:2])
                assert (major, minor) >= (3, 10), (
                    f"[tool.mypy] python_version = '{version_str}' is too old for "
                    f"types.EllipsisType (need >= 3.10). Update to '3.10'."
                )
                return
        # If we reach here, python_version was not found in [tool.mypy]
        raise AssertionError(
            "python_version not found in [tool.mypy] section of pyproject.toml"
        )



"""TDD cycle 3: 失敗するテストを書いてから修正する。

対応した問題（コードレビューコメント from 1a9d4e4）:
1. TestNoUnusedImports が ruff 未インストール環境でエラーになる
   → ruff が利用できない場合は pytest.skip すべき
2. docs/en/api/nanasqlite.md の table() シグネチャに cache_ttl / cache_persistence_ttl がない
3. docs/ja/api/nanasqlite.md の table() シグネチャに cache_ttl / cache_persistence_ttl がない
"""
from __future__ import annotations

from pathlib import Path

REPO_ROOT = Path(__file__).parent.parent


def get_table_signature_block(doc_path: Path) -> str:
    """指定された Markdown ドキュメントから table() シグネチャのコードブロックを取得する。

    python コードブロック（バッククォート3つ + python）の中で "def table(" を含むブロックを探し、
    そのブロックの内容（区切り文字を除いたコード部分）を返す。
    一致するブロックが見つからない場合は空文字列を返す。
    """
    content = doc_path.read_text(encoding="utf-8")
    lines = content.splitlines()
    in_python_block = False
    in_table_block = False
    block_lines: list[str] = []
    for line in lines:
        stripped = line.strip()
        if stripped.startswith("```python"):
            in_python_block = True
            in_table_block = False
            block_lines = []
            continue
        if stripped == "```" and in_python_block:
            if in_table_block:
                # table() ブロックが終了した — 収集完了
                return "\n".join(block_lines)
            # table() を含まないブロックが終わった — 状態をリセット
            in_python_block = False
            in_table_block = False
            block_lines = []
            continue
        if in_python_block:
            block_lines.append(line)
            if "def table(" in line:
                in_table_block = True
    return "\n".join(block_lines) if in_table_block else ""


class TestRuffAvailabilitySkip:
    """TestNoUnusedImports が ruff 未利用時にスキップすることを確認する。"""

    def test_tdd_cycle_2_ruff_check_handles_unavailable(self):
        """test_tdd_cycle_2.py の TestNoUnusedImports が ruff 未インストール時に対応していること。

        subprocess が 'No module named ruff' 相当のエラーを返した場合に
        テストが誤判定しないことを確認する。
        """
        src = (REPO_ROOT / "tests" / "test_tdd_cycle_2.py").read_text(encoding="utf-8")
        # TestNoUnusedImports クラス部分を抽出する
        class_start = src.find("class TestNoUnusedImports:")
        assert class_start != -1, "TestNoUnusedImports クラスが test_tdd_cycle_2.py に見つからない"
        class_src = src[class_start:]
        # ruff が利用できるかチェックしているか確認する
        has_availability_check = (
            "shutil.which" in class_src
            or "pytest.skip" in class_src
            or "importorskip" in class_src
        )
        assert has_availability_check, (
            "TestNoUnusedImports は ruff が利用できない場合の処理"
            "（shutil.which + pytest.skip など）を含む必要があります。"
            " ruff が CI にインストールされていない場合にテストが誤って失敗します。"
        )


class TestTableSignatureInEnDocs:
    """docs/en/api/nanasqlite.md の table() シグネチャに cache_ttl/cache_persistence_ttl が含まれること。"""

    def test_en_docs_table_signature_has_cache_ttl(self):
        """docs/en/api/nanasqlite.md の table() シグネチャに cache_ttl が含まれること。"""
        doc_path = REPO_ROOT / "docs" / "en" / "api" / "nanasqlite.md"
        block = get_table_signature_block(doc_path)
        assert "cache_ttl" in block, (
            "docs/en/api/nanasqlite.md の table() シグネチャに cache_ttl が含まれていません。"
            f"\n見つかったブロック:\n{block}"
        )

    def test_en_docs_table_signature_has_cache_persistence_ttl(self):
        """docs/en/api/nanasqlite.md の table() シグネチャに cache_persistence_ttl が含まれること。"""
        doc_path = REPO_ROOT / "docs" / "en" / "api" / "nanasqlite.md"
        block = get_table_signature_block(doc_path)
        assert "cache_persistence_ttl" in block, (
            "docs/en/api/nanasqlite.md の table() シグネチャに cache_persistence_ttl が含まれていません。"
            f"\n見つかったブロック:\n{block}"
        )


class TestTableSignatureInJaDocs:
    """docs/ja/api/nanasqlite.md の table() シグネチャに cache_ttl/cache_persistence_ttl が含まれること。"""

    def test_ja_docs_table_signature_has_cache_ttl(self):
        """docs/ja/api/nanasqlite.md の table() シグネチャに cache_ttl が含まれること。"""
        doc_path = REPO_ROOT / "docs" / "ja" / "api" / "nanasqlite.md"
        block = get_table_signature_block(doc_path)
        assert "cache_ttl" in block, (
            "docs/ja/api/nanasqlite.md の table() シグネチャに cache_ttl が含まれていません。"
            f"\n見つかったブロック:\n{block}"
        )

    def test_ja_docs_table_signature_has_cache_persistence_ttl(self):
        """docs/ja/api/nanasqlite.md の table() シグネチャに cache_persistence_ttl が含まれること。"""
        doc_path = REPO_ROOT / "docs" / "ja" / "api" / "nanasqlite.md"
        block = get_table_signature_block(doc_path)
        assert "cache_persistence_ttl" in block, (
            "docs/ja/api/nanasqlite.md の table() シグネチャに cache_persistence_ttl が含まれていません。"
            f"\n見つかったブロック:\n{block}"
        )




"""
validkit-py オプション連携のテスト
(Tests for optional validkit-py integration)
"""
import pytest

from nanasqlite import HAS_VALIDKIT, AsyncNanaSQLite, NanaSQLite, NanaSQLiteValidationError
from nanasqlite import core as core_mod

validkit_installed = HAS_VALIDKIT

if validkit_installed:
    from validkit import v
else:
    v = None  # type: ignore[assignment]


def test_has_validkit_flag():
    """HAS_VALIDKIT フラグが正しく設定されているか確認する。"""
    assert hasattr(core_mod, "HAS_VALIDKIT")
    assert isinstance(core_mod.HAS_VALIDKIT, bool)


def test_has_validkit_flag_from_package():
    """HAS_VALIDKIT が nanasqlite パッケージから直接インポートできることを確認する。"""
    assert isinstance(HAS_VALIDKIT, bool)
    assert HAS_VALIDKIT is validkit_installed


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_validator_accepts_valid_value(tmp_path):
    """バリデーションスキーマに一致する値は正常に保存される。"""
    schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "valid.db"), validator=schema)

    db["user"] = {"name": "Alice", "age": 30}
    assert db["user"] == {"name": "Alice", "age": 30}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_validator_rejects_invalid_value(tmp_path):
    """バリデーションスキーマに違反する値を設定すると NanaSQLiteValidationError が送出される。"""
    schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "invalid.db"), validator=schema)

    with pytest.raises(NanaSQLiteValidationError):
        db["user"] = {"name": "Bob", "age": "not_an_int"}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_validator_does_not_affect_db_when_invalid(tmp_path):
    """バリデーションエラー時はDBに値が保存されないことを確認する。"""
    schema = {"name": v.str(), "score": v.int()}
    db = NanaSQLite(str(tmp_path / "no_write.db"), validator=schema)

    with pytest.raises(NanaSQLiteValidationError):
        db["entry"] = {"name": "Carol", "score": "oops"}

    assert "entry" not in db


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_no_validator_allows_any_value(tmp_path):
    """validator を指定しない場合は任意の値を保存できる。"""
    db = NanaSQLite(str(tmp_path / "novalidator.db"))

    db["misc"] = {"anything": [1, 2, 3], "flag": True}
    assert db["misc"]["flag"] is True


@pytest.mark.skipif(validkit_installed, reason="このテストは validkit-py 未インストール環境のみ実行します")
def test_validator_raises_import_error_when_validkit_missing(tmp_path):
    """validkit-py が未インストールの状態で validator を渡すと ImportError が送出される。"""
    with pytest.raises(ImportError, match="validkit-py"):
        NanaSQLite(str(tmp_path / "err.db"), validator={"key": "value"})


@pytest.mark.skipif(validkit_installed, reason="このテストは validkit-py 未インストール環境のみ実行します")
@pytest.mark.asyncio
async def test_async_validator_raises_import_error_when_validkit_missing(tmp_path):
    """validkit-py が未インストールの状態で AsyncNanaSQLite に validator を渡すと ImportError が送出される。"""
    with pytest.raises(ImportError, match="validkit-py"):
        AsyncNanaSQLite(str(tmp_path / "async_err.db"), validator={"key": "value"})


# ========================== table() per-table validator tests ==========================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_table_inherits_parent_validator(tmp_path):
    """table() で validator を指定しない場合、親の validator が引き継がれる。"""
    schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "inherit.db"), validator=schema)
    users_db = db.table("users")

    # 子テーブルでも有効な値は保存できる
    users_db["u1"] = {"name": "Dave", "age": 25}
    assert users_db["u1"] == {"name": "Dave", "age": 25}

    # 子テーブルでもスキーマ違反は拒否される
    with pytest.raises(NanaSQLiteValidationError):
        users_db["u2"] = {"name": "Eve", "age": "bad"}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_table_override_validator(tmp_path):
    """table() で独自の validator を渡すと、親のスキーマではなくそのスキーマが使われる。"""
    parent_schema = {"x": v.int()}
    child_schema = {"name": v.str(), "score": v.float()}
    db = NanaSQLite(str(tmp_path / "override.db"), validator=parent_schema)
    scores_db = db.table("scores", validator=child_schema)

    # 子スキーマ準拠の値は保存できる
    scores_db["s1"] = {"name": "Frank", "score": 9.5}
    assert scores_db["s1"] == {"name": "Frank", "score": 9.5}

    # 子スキーマ違反は拒否される
    with pytest.raises(NanaSQLiteValidationError):
        scores_db["s2"] = {"name": "Grace", "score": "not_a_float"}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_table_no_parent_validator_no_child_validator(tmp_path):
    """親も子も validator なしの場合、任意の値を書き込める。"""
    db = NanaSQLite(str(tmp_path / "novalidator.db"))
    child_db = db.table("misc")

    child_db["k"] = {"anything": True, "nums": [1, 2]}
    assert child_db["k"]["anything"] is True


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_table_explicit_none_disables_parent_validator(tmp_path):
    """table(validator=None) を明示すると親のスキーマを無効化して任意値を書き込める。"""
    parent_schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "disable.db"), validator=parent_schema)
    # validator=None を明示すると親スキーマを引き継がない
    free_db = db.table("free_table", validator=None)

    # バリデーションなしなので任意の値を書き込める
    free_db["k"] = {"anything": True, "n": 99}
    assert free_db["k"]["anything"] is True


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_table_child_does_not_affect_parent_validator(tmp_path):
    """table() で別スキーマを持たせても親インスタンスのバリデーションは変わらない。"""
    parent_schema = {"x": v.int()}
    child_schema = {"name": v.str()}
    db = NanaSQLite(str(tmp_path / "separate.db"), validator=parent_schema)
    child_db = db.table("child", validator=child_schema)

    # 親テーブルは親スキーマで検証される
    db["valid"] = {"x": 42}
    with pytest.raises(NanaSQLiteValidationError):
        db["bad"] = {"x": "not_int"}

    # 子テーブルは子スキーマで検証される
    child_db["c1"] = {"name": "Helen"}
    with pytest.raises(NanaSQLiteValidationError):
        child_db["c2"] = {"name": 999}


# ========================== AsyncNanaSQLite validator tests ==========================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_validator_accepts_valid_value(tmp_path):
    """AsyncNanaSQLite: バリデーションスキーマに一致する値は正常に保存される。"""
    schema = {"name": v.str(), "age": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_valid.db"), validator=schema) as db:
        await db.aset("user", {"name": "Alice", "age": 30})
        assert await db.aget("user") == {"name": "Alice", "age": 30}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_validator_rejects_invalid_value(tmp_path):
    """AsyncNanaSQLite: バリデーションスキーマに違反する値は NanaSQLiteValidationError が送出される。"""
    schema = {"name": v.str(), "age": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_invalid.db"), validator=schema) as db:
        with pytest.raises(NanaSQLiteValidationError):
            await db.aset("user", {"name": "Bob", "age": "not_an_int"})


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_validator_does_not_affect_db_when_invalid(tmp_path):
    """AsyncNanaSQLite: バリデーションエラー時はDBに値が保存されないことを確認する。"""
    schema = {"name": v.str(), "score": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_no_write.db"), validator=schema) as db:
        with pytest.raises(NanaSQLiteValidationError):
            await db.aset("entry", {"name": "Carol", "score": "oops"})
        assert not await db.acontains("entry")


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_table_inherits_parent_validator(tmp_path):
    """AsyncNanaSQLite: table() で validator を指定しない場合、親の validator が引き継がれる。"""
    schema = {"name": v.str(), "age": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_inherit.db"), validator=schema) as db:
        users_db = await db.table("users")

        await users_db.aset("u1", {"name": "Dave", "age": 25})
        assert await users_db.aget("u1") == {"name": "Dave", "age": 25}

        with pytest.raises(NanaSQLiteValidationError):
            await users_db.aset("u2", {"name": "Eve", "age": "bad"})


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_table_override_validator(tmp_path):
    """AsyncNanaSQLite: table() で独自の validator を渡すと、そのスキーマが使われる。"""
    parent_schema = {"x": v.int()}
    child_schema = {"name": v.str(), "score": v.float()}
    async with AsyncNanaSQLite(str(tmp_path / "async_override.db"), validator=parent_schema) as db:
        scores_db = await db.table("scores", validator=child_schema)

        await scores_db.aset("s1", {"name": "Frank", "score": 9.5})
        assert await scores_db.aget("s1") == {"name": "Frank", "score": 9.5}

        with pytest.raises(NanaSQLiteValidationError):
            await scores_db.aset("s2", {"name": "Grace", "score": "not_a_float"})


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_table_explicit_none_disables_parent_validator(tmp_path):
    """AsyncNanaSQLite: table(validator=None) を明示すると親のスキーマを無効化できる。"""
    parent_schema = {"name": v.str(), "age": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_disable.db"), validator=parent_schema) as db:
        free_db = await db.table("free_table", validator=None)

        # バリデーションなしなので任意の値を書き込める
        await free_db.aset("k", {"anything": True, "n": 99})
        assert (await free_db.aget("k"))["anything"] is True


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_no_validator_allows_any_value(tmp_path):
    """AsyncNanaSQLite: validator を指定しない場合は任意の値を保存できる。"""
    async with AsyncNanaSQLite(str(tmp_path / "async_novalidator.db")) as db:
        await db.aset("misc", {"anything": [1, 2, 3], "flag": True})
        result = await db.aget("misc")
        assert result["flag"] is True


# ========================== batch_update() validation tests ==========================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_batch_update_validates_all_before_writing(tmp_path):
    """batch_update() はすべての値をDB書き込み前に検証し、1件でも違反があれば何も書かない。"""
    schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "batch.db"), validator=schema)

    mapping_with_invalid = {
        "u1": {"name": "Alice", "age": 30},
        "u2": {"name": "Bob", "age": "bad"},  # 不正な値
    }

    with pytest.raises(NanaSQLiteValidationError):
        db.batch_update(mapping_with_invalid)

    # 1件でも失敗した場合、全件DBに書き込まれない（アトミックな失敗保証）
    assert "u1" not in db
    assert "u2" not in db


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_batch_update_all_valid_writes_all(tmp_path):
    """batch_update() は全値がバリデーションを通過すれば全件書き込む。"""
    schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "batch_valid.db"), validator=schema)

    db.batch_update({
        "u1": {"name": "Alice", "age": 30},
        "u2": {"name": "Bob", "age": 25},
    })

    assert db["u1"] == {"name": "Alice", "age": 30}
    assert db["u2"] == {"name": "Bob", "age": 25}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_batch_update_validates_all_before_writing(tmp_path):
    """AsyncNanaSQLite: batch_update() は1件でも違反があれば何も書かない。"""
    schema = {"name": v.str(), "age": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_batch.db"), validator=schema) as db:
        mapping_with_invalid = {
            "u1": {"name": "Alice", "age": 30},
            "u2": {"name": "Bob", "age": "bad"},
        }

        with pytest.raises(NanaSQLiteValidationError):
            await db.abatch_update(mapping_with_invalid)

        assert not await db.acontains("u1")
        assert not await db.acontains("u2")


# ========================== coerce mode tests ==========================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_coerce_converts_value_on_setitem(tmp_path):
    """coerce=True のとき、__setitem__ が変換済みの値を保存する。"""
    schema = {"age": v.int().coerce()}
    db = NanaSQLite(str(tmp_path / "coerce.db"), validator=schema, coerce=True)

    db["user"] = {"age": "42"}  # 文字列 → int に自動変換
    assert db["user"] == {"age": 42}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_no_coerce_keeps_original_value(tmp_path):
    """coerce=False（デフォルト）のとき、バリデーションは通っても元の値がそのまま保存される。"""
    schema = {"age": v.int()}
    db = NanaSQLite(str(tmp_path / "no_coerce.db"), validator=schema)

    db["user"] = {"age": 42}
    assert db["user"] == {"age": 42}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_coerce_in_batch_update(tmp_path):
    """coerce=True のとき、batch_update() も変換済みの値を保存する。"""
    schema = {"score": v.float().coerce()}
    db = NanaSQLite(str(tmp_path / "batch_coerce.db"), validator=schema, coerce=True)

    db.batch_update({
        "a": {"score": "1.5"},
        "b": {"score": "2.0"},
    })

    assert db["a"] == {"score": 1.5}
    assert db["b"] == {"score": 2.0}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_coerce_inherited_by_table(tmp_path):
    """coerce=True は table() で子テーブルに引き継がれる。"""
    schema = {"age": v.int().coerce()}
    db = NanaSQLite(str(tmp_path / "coerce_inherit.db"), validator=schema, coerce=True)
    child_db = db.table("child")

    child_db["c1"] = {"age": "7"}
    assert child_db["c1"] == {"age": 7}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_coerce_can_be_disabled_per_table(tmp_path):
    """table(coerce=False) で子テーブルの自動変換を無効にできる。"""
    schema = {"age": v.int()}
    db = NanaSQLite(str(tmp_path / "coerce_override.db"), validator=schema, coerce=True)
    child_db = db.table("child", coerce=False)

    # 子テーブルは coerce=False なので、変換なしで型が一致する値のみ受け付ける
    child_db["c1"] = {"age": 5}
    assert child_db["c1"] == {"age": 5}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_coerce_converts_value(tmp_path):
    """AsyncNanaSQLite: coerce=True のとき自動変換が有効になる。"""
    schema = {"age": v.int().coerce()}
    async with AsyncNanaSQLite(str(tmp_path / "async_coerce.db"), validator=schema, coerce=True) as db:
        await db.aset("user", {"age": "99"})
        assert await db.aget("user") == {"age": 99}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_coerce_inherited_by_table(tmp_path):
    """AsyncNanaSQLite: coerce=True は table() で子テーブルに引き継がれる。"""
    schema = {"age": v.int().coerce()}
    async with AsyncNanaSQLite(str(tmp_path / "async_coerce_inherit.db"), validator=schema, coerce=True) as db:
        child_db = await db.table("child")
        await child_db.aset("c1", {"age": "3"})
        assert await child_db.aget("c1") == {"age": 3}


# ========================== coerce dual-requirement tests ==========================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_coerce_true_without_field_coerce_still_fails(tmp_path):
    """coerce=True on NanaSQLite alone does NOT convert types if the field validator lacks .coerce().

    Both coerce=True (NanaSQLite) AND .coerce() on the field validator are needed.
    Without .coerce() on the field, type mismatches still raise NanaSQLiteValidationError.
    """
    # Schema without field .coerce()
    schema = {"age": v.int()}
    db = NanaSQLite(str(tmp_path / "dual_req.db"), validator=schema, coerce=True)

    # Wrong type input must still fail - coerce=True on NanaSQLite alone is not enough
    with pytest.raises(NanaSQLiteValidationError):
        db["user"] = {"age": "30"}  # str, not int - no field .coerce()

    # Correct type passes as before
    db["user"] = {"age": 30}
    assert db["user"] == {"age": 30}


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_field_coerce_without_nanasqlite_coerce_stores_original(tmp_path):
    """Field .coerce() without coerce=True on NanaSQLite: validation passes, original value stored.

    When the field has .coerce(), validkit accepts the input (converting internally to validate),
    but NanaSQLite stores the original value because coerce=False (default).
    """
    schema = {"age": v.int().coerce()}
    db = NanaSQLite(str(tmp_path / "field_coerce_only.db"), validator=schema, coerce=False)

    # Validation passes (validkit coerces internally), original string is stored
    db["user"] = {"age": "30"}
    assert db["user"] == {"age": "30"}  # original value, NOT converted


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
def test_batch_update_partial_writes_valid_values_and_reports_failures(tmp_path):
    """batch_update_partial() は違反キーだけを返し、正常値は保存する。"""
    schema = {"name": v.str(), "age": v.int()}
    db = NanaSQLite(str(tmp_path / "batch_partial.db"), validator=schema)

    failed = db.batch_update_partial({
        "u1": {"name": "Alice", "age": 30},
        "u2": {"name": "Bob", "age": "bad"},
        "u3": {"name": "Carol", "age": 22},
    })

    assert set(failed) == {"u2"}
    assert "failed schema validation" in failed["u2"]
    assert db["u1"] == {"name": "Alice", "age": 30}
    assert db["u3"] == {"name": "Carol", "age": 22}
    assert "u2" not in db


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
@pytest.mark.asyncio
async def test_async_batch_update_partial_writes_valid_values_and_reports_failures(tmp_path):
    """AsyncNanaSQLite: batch_update_partial() は違反キーだけを返し、正常値は保存する。"""
    schema = {"name": v.str(), "age": v.int()}
    async with AsyncNanaSQLite(str(tmp_path / "async_batch_partial.db"), validator=schema) as db:
        failed = await db.batch_update_partial({
            "u1": {"name": "Alice", "age": 30},
            "u2": {"name": "Bob", "age": "bad"},
            "u3": {"name": "Carol", "age": 22},
        })

        assert set(failed) == {"u2"}
        assert await db.aget("u1") == {"name": "Alice", "age": 30}
        assert await db.aget("u3") == {"name": "Carol", "age": 22}
        assert not await db.acontains("u2")



import pytest

from nanasqlite import AsyncNanaSQLite, NanaSQLite

# ==================== Synchronous Edge Cases ====================


class TestSyncEdgeCases:
    @pytest.fixture
    def db(self, tmp_path):
        db_path = str(tmp_path / "edge_cases_sync.db")
        conn = NanaSQLite(db_path)
        yield conn
        conn.close()

    def test_batch_operations_empty(self, db):
        """Verify behavior with empty inputs for batch operations."""
        # batch_get with empty list -> empty dict
        assert db.batch_get([]) == {}

        # batch_update with empty dict -> no error
        db.batch_update({})

        # batch_delete with empty list -> no error
        db.batch_delete([])

    def test_none_value_storage(self, tmp_path):
        """Verify that storing None (null) does not cause data loss upon reload."""
        db_path = str(tmp_path / "edge_cases_none.db")
        db = NanaSQLite(db_path)
        try:
            db["test_none"] = None
            assert db["test_none"] is None
            assert "test_none" in db
        finally:
            db.close()

        # Reopen database to check if None value is retained and not causing KeyError
        db2 = NanaSQLite(db_path)
        try:
            assert "test_none" in db2
            assert db2["test_none"] is None
        finally:
            db2.close()

    def test_pagination_edge_cases(self, db):
        """Verify pagination edge cases."""
        # Setup data
        db.batch_update({f"k{i}": i for i in range(10)})
        table = "data"  # default table

        # limit=0 -> empty list
        results = db.query_with_pagination(table_name=table, limit=0)
        assert len(results) == 0

        # offset > total -> empty list
        results = db.query_with_pagination(table_name=table, limit=5, offset=100)
        assert len(results) == 0

        # offset=0 -> start from beginning
        results = db.query_with_pagination(table_name=table, limit=5, offset=0)
        assert len(results) == 5

    def test_pagination_negative_values(self, db):
        """Verify negative values in pagination (Validation Check)."""
        # NanaSQLite explicitly invalidates negative limits
        db.batch_update({"a": 1, "b": 2})

        with pytest.raises(ValueError, match="limit must be non-negative"):
            db.query_with_pagination(limit=-1)


# ==================== Asynchronous Edge Cases ====================


class TestAsyncEdgeCases:
    @pytest.fixture
    def db_path(self, tmp_path):
        return str(tmp_path / "edge_cases_async.db")

    @pytest.mark.asyncio
    async def test_async_batch_operations_empty(self, db_path):
        """Verify async behavior with empty inputs."""
        async with AsyncNanaSQLite(db_path) as db:
            # abatch_get
            res = await db.abatch_get([])
            assert res == {}

            # abatch_update
            await db.abatch_update({})

            # abatch_delete
            await db.abatch_delete([])

    @pytest.mark.asyncio
    async def test_async_pagination_edge_cases(self, db_path):
        """Verify async pagination edge cases."""
        async with AsyncNanaSQLite(db_path) as db:
            # Setup
            await db.abatch_update({f"k{i}": i for i in range(10)})

            # limit=0
            results = await db.aquery_with_pagination(limit=0)
            assert len(results) == 0

            # offset > total
            results = await db.aquery_with_pagination(limit=5, offset=100)
            assert len(results) == 0

    @pytest.mark.asyncio
    async def test_async_query_empty_conditions(self, db_path):
        """Verify query behavior with minimal arguments."""
        async with AsyncNanaSQLite(db_path) as db:
            await db.aset("key", "value")

            # No args -> essentially 'SELECT *' (depends on implementation defaults)
            # Core implementation of `query` constructs SELECT specific columns or *
            results = await db.aquery()
            assert len(results) > 0
            assert results[0]["key"] == "key"



"""
Tests for the cache strategy feature (v1.3.0).
"""

import pytest

from nanasqlite import CacheType, NanaSQLite


class TestCacheType:
    """Test CacheType enum."""

    def test_cache_type_values(self):
        assert CacheType.UNBOUNDED.value == "unbounded"
        assert CacheType.LRU.value == "lru"


class TestUnboundedCache:
    """Test Unbounded (default) cache strategy."""

    def test_default_is_unbounded(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path)) as db:
            # Default should be unbounded
            db["key1"] = "value1"
            assert db["key1"] == "value1"

    def test_explicit_unbounded(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.UNBOUNDED) as db:
            for i in range(100):
                db[f"key{i}"] = f"value{i}"
            # All 100 keys should be in cache (unbounded)
            assert db._cache.size == 100


class TestLRUCache:
    """Test LRU cache strategy."""

    def test_lru_requires_size(self, tmp_path):
        db_path = tmp_path / "test.db"
        with pytest.raises(ValueError, match="cache_size must be"):
            NanaSQLite(str(db_path), cache_strategy=CacheType.LRU)

    def test_lru_basic_operations(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.LRU, cache_size=10) as db:
            db["a"] = 1
            db["b"] = 2
            assert db["a"] == 1
            assert db["b"] == 2

    def test_lru_eviction(self, tmp_path):
        """Test that LRU evicts oldest entries when full."""
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.LRU, cache_size=3) as db:
            # Insert 5 items into cache of size 3
            for i in range(5):
                db[f"key{i}"] = f"value{i}"

            # Cache should only have 3 items (key2, key3, key4)
            assert db._cache.size == 3
            cache_keys = list(db._cache.get_data().keys())
            assert "key2" in cache_keys
            assert "key3" in cache_keys
            assert "key4" in cache_keys
            # key0 and key1 should be evicted
            assert "key0" not in cache_keys
            assert "key1" not in cache_keys

    def test_lru_db_persistence(self, tmp_path):
        """Test that evicted items are still in DB."""
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.LRU, cache_size=2) as db:
            db["a"] = 1
            db["b"] = 2
            db["c"] = 3  # This evicts 'a' from cache

            # 'a' is not in cache but should be in DB
            assert "a" not in db._cache.get_data()
            # Accessing 'a' should reload from DB
            assert db["a"] == 1

    def test_lru_access_updates_order(self, tmp_path):
        """Test that accessing an item moves it to end (most recently used)."""
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.LRU, cache_size=3) as db:
            db["a"] = 1
            db["b"] = 2
            db["c"] = 3
            # Access 'a' to make it recently used
            _ = db["a"]
            # Add new item, should evict 'b' (oldest)
            db["d"] = 4
            cache_keys = list(db._cache.get_data().keys())
            assert "b" not in cache_keys
            assert "a" in cache_keys


class TestTableCacheStrategy:
    """Test cache strategy on sub-tables."""

    def test_table_inherits_default(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path)) as db:
            sub = db.table("users")
            sub["user1"] = {"name": "Alice"}
            assert sub["user1"]["name"] == "Alice"

    def test_table_custom_strategy(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path)) as db:
            # Main db is unbounded
            logs = db.table("logs", cache_strategy=CacheType.LRU, cache_size=5)
            for i in range(10):
                logs[f"log{i}"] = f"message{i}"
            # Logs table should only cache 5 items
            assert logs._cache.size == 5


class TestCacheClearAndRefresh:
    """Test cache clearing and refreshing."""

    def test_clear_cache(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.LRU, cache_size=10) as db:
            db["a"] = 1
            db["b"] = 2
            assert db._cache.size == 2
            db.clear_cache()
            assert db._cache.size == 0
            # Data should still be in DB
            assert db["a"] == 1

    def test_refresh(self, tmp_path):
        db_path = tmp_path / "test.db"
        with NanaSQLite(str(db_path), cache_strategy=CacheType.LRU, cache_size=10) as db:
            db["a"] = 1
            # Manually modify cache for testing
            db._cache.set("a", 999)
            assert db["a"] == 999  # Cache hit
            # Refresh should reload from DB
            db.refresh("a")
            assert db["a"] == 1



"""
NanaSQLite Cache Module: Provides multiple cache strategy implementations.
(NanaSQLite用キャッシュモジュール: 複数のキャッシュ戦略実装を提供)
"""
from __future__ import annotations

import logging
from abc import abstractmethod
from collections import OrderedDict
from collections.abc import MutableMapping
from enum import Enum
from typing import Any, Callable, Protocol

from .utils import ExpiringDict

logger = logging.getLogger(__name__)

# Try to import C-optimized LRU (optional dependency)
_FAST_LRU: Any = None
try:
    from lru import LRU

    _FAST_LRU = LRU
    HAS_FAST_LRU = True
except ImportError:
    HAS_FAST_LRU = False


class CacheType(str, Enum):
    """Available cache strategies"""

    UNBOUNDED = "unbounded"
    LRU = "lru"
    TTL = "ttl"


class CacheStrategy(Protocol):
    """Protocol defining the interface for all cache implementations"""

    @abstractmethod
    def get(self, key: str) -> Any | None:
        """Get item from cache. Returns None if not found."""

    @abstractmethod
    def set(self, key: str, value: Any) -> None:
        """Set item in cache."""

    @abstractmethod
    def delete(self, key: str) -> None:
        """Remove item from cache."""

    @abstractmethod
    def invalidate(self, key: str) -> None:
        """Completely remove key from cache knowledge (forget it exists or not)."""

    @abstractmethod
    def contains(self, key: str) -> bool:
        """Check if key exists in cache."""

    @abstractmethod
    def mark_cached(self, key: str) -> None:
        """Mark a key as 'known' (cached) even if value is not loaded yet."""

    @abstractmethod
    def is_cached(self, key: str) -> bool:
        """Check if we have knowledge about this key (either value or existence)."""

    @abstractmethod
    def clear(self) -> None:
        """Clear entire cache."""

    @abstractmethod
    def get_data(self) -> MutableMapping[str, Any]:
        """Return reference to internal data storage (for legacy compatibility/inspection)."""

    @property
    @abstractmethod
    def size(self) -> int:
        """Current number of items in cache."""


class UnboundedCache(CacheStrategy):
    """
    Default behavior: Infinite growth, maximum speed.
    Equivalent to v1.2.x logic using standard user Dict + Set.
    If max_size is set, uses FIFO eviction.
    """

    def __init__(self, max_size: int | None = None):
        self._max_size = max_size
        self._data: dict[str, Any] = {}
        self._cached_keys: set[str] = set()

    def get(self, key: str) -> Any | None:
        return self._data.get(key)

    def set(self, key: str, value: Any) -> None:
        if self._max_size and self._max_size > 0:
            if key not in self._data and len(self._data) >= self._max_size:
                # FIFO:最初に追加された要素を削除 (Python 3.7+)
                oldest_key = next(iter(self._data))
                del self._data[oldest_key]
                self._cached_keys.discard(oldest_key)

        self._data[key] = value
        self._cached_keys.add(key)

    def delete(self, key: str) -> None:
        if key in self._data:
            del self._data[key]
        self._cached_keys.add(key)

    def invalidate(self, key: str) -> None:
        if key in self._data:
            del self._data[key]
        self._cached_keys.discard(key)

    def contains(self, key: str) -> bool:
        return key in self._data

    def mark_cached(self, key: str) -> None:
        self._cached_keys.add(key)

    def is_cached(self, key: str) -> bool:
        return key in self._cached_keys

    def clear(self) -> None:
        self._data.clear()
        self._cached_keys.clear()

    def get_data(self) -> MutableMapping[str, Any]:
        return self._data

    @property
    def size(self) -> int:
        return len(self._data)


class StdLRUCache(CacheStrategy):
    """
    Standard Library LRU implementation using OrderedDict.
    Safe fallback if C extensions are not available.
    """

    def __init__(self, max_size: int):
        self._max_size = max_size
        self._data: OrderedDict[str, Any] = OrderedDict()
        # For LRU, is_cached logic is unified with containment.
        # If it's evicted, it's no longer "cached", so we must fetch again.

    def get(self, key: str) -> Any | None:
        if key not in self._data:
            return None
        self._data.move_to_end(key)  # Mark used
        return self._data[key]

    def set(self, key: str, value: Any) -> None:
        self._data[key] = value
        self._data.move_to_end(key)
        if len(self._data) > self._max_size:
            self._data.popitem(last=False)  # Evict oldest

    def delete(self, key: str) -> None:
        if key in self._data:
            del self._data[key]

    def invalidate(self, key: str) -> None:
        self.delete(key)

    def contains(self, key: str) -> bool:
        return key in self._data

    def mark_cached(self, key: str) -> None:
        # In strict LRU, we can't just "mark" likely.
        # But to support the logic of "key is known to exist",
        # we might need to store it. For now, we only store actual values.
        pass

    def is_cached(self, key: str) -> bool:
        # In LRU mode, if it's in the map, it's cached.
        # If it's not, we consider it "not cached" (needs DB fetch)
        return key in self._data

    def clear(self) -> None:
        self._data.clear()

    def get_data(self) -> MutableMapping[str, Any]:
        return self._data

    @property
    def size(self) -> int:
        return len(self._data)


class FastLRUCache(CacheStrategy):
    """
    High-performance LRU implementation using lru-dict (C extension).
    """

    def __init__(self, max_size: int):
        if not HAS_FAST_LRU or _FAST_LRU is None:
            raise ImportError("lru-dict is not installed. Use 'fast_lru' extra or fallback to StdLRUCache.")
        self._data = _FAST_LRU(max_size)

    def get(self, key: str) -> Any | None:
        return self._data.get(key)

    def set(self, key: str, value: Any) -> None:
        self._data[key] = value

    def delete(self, key: str) -> None:
        if key in self._data:
            del self._data[key]

    def invalidate(self, key: str) -> None:
        self.delete(key)

    def contains(self, key: str) -> bool:
        return key in self._data

    def mark_cached(self, key: str) -> None:
        pass

    def is_cached(self, key: str) -> bool:
        return key in self._data

    def clear(self) -> None:
        self._data.clear()

    def get_data(self) -> MutableMapping[str, Any]:
        # lru-dict behaves like a dict
        return self._data  # type: ignore

    @property
    def size(self) -> int:
        return len(self._data)


class TTLCache(CacheStrategy):
    """
    有効期限付きキャッシュ (TTL)
    """

    def __init__(
        self,
        ttl: float,
        max_size: int | None = None,
        on_expire: Callable[[str, Any], None] | None = None,
    ):
        """
        Args:
            ttl: 有効期限（秒）
            max_size: 最大保持件数 (FIFO併用、None の場合は無制限)
            on_expire: 有効期限切れ時のコールバック
        """

        self._data = ExpiringDict(expiration_time=ttl, on_expire=on_expire)
        self._max_size = max_size
        self._cached_keys: set[str] = set()

    def get(self, key: str) -> Any | None:
        return self._data.get(key)

    def set(self, key: str, value: Any) -> None:
        if self._max_size and self._max_size > 0:
            if key not in self._data and len(self._data) >= self._max_size:
                oldest_key = next(iter(self._data))
                del self._data[oldest_key]
                self._cached_keys.discard(oldest_key)

        self._data[key] = value
        self._cached_keys.add(key)

    def delete(self, key: str) -> None:
        if key in self._data:
            del self._data[key]
        self._cached_keys.discard(key)

    def clear(self) -> None:
        self._data.clear()
        self._cached_keys.clear()

    def mark_cached(self, key: str) -> None:
        self._cached_keys.add(key)

    def is_cached(self, key: str) -> bool:
        return key in self._data or key in self._cached_keys

    def invalidate(self, key: str) -> None:
        self.delete(key)

    def contains(self, key: str) -> bool:
        return key in self._data

    @property
    def size(self) -> int:
        return len(self._data)

    def get_data(self) -> MutableMapping[str, Any]:
        return self._data


def create_cache(
    strategy: str | CacheType = CacheType.UNBOUNDED,
    size: int | None = None,
    ttl: float | None = None,
    on_expire: Callable[[str, Any], None] | None = None,
) -> CacheStrategy:
    """Factory to create appropriate cache instance"""
    # Normalize strategy
    if isinstance(strategy, CacheType):
        strategy = strategy.value

    if strategy == CacheType.LRU:
        if size is None or size <= 0:
            raise ValueError("cache_size must be a positive integer when using LRU strategy")

        if HAS_FAST_LRU:
            logger.info("Using FastLRUCache (lru-dict) with size %s", size)
            return FastLRUCache(size)
        logger.warning(
            "lru-dict not found. Falling back to standard LRUCache (OrderedDict) with size %s", size
        )
        return StdLRUCache(size)

    if strategy == CacheType.TTL:
        if ttl is None or ttl <= 0:
            raise ValueError("cache_ttl must be a positive value when using TTL strategy")
        return TTLCache(ttl, max_size=size, on_expire=on_expire)

    return UnboundedCache(max_size=size)



"""TDD cycle 2: 失敗するテストを書いてから修正する。

対応した問題（コードレビューコメント from 54c4a16）:
1. NanaSQLite.__init__ の validator デフォルトが _UNSET → None に変更すべき
   (コンストラクターに「親から継承」の概念はないため)
2. table() に cache_ttl / cache_persistence_ttl パラメーターを追加すべき
   (親が非 TTL の場合でも cache_strategy=TTL で table() を呼べるよう)
3. test_table_inheritance_comprehensive.py に未使用 `import time` (Ruff F401)
4. test_tdd_review_fixes.py に未使用 `import pytest` (Ruff F401)
5. test_table_inheritance_comprehensive.py に未使用 `NanaSQLiteValidationError` (Ruff F401)
"""
from __future__ import annotations

import inspect
import shutil
import subprocess
import sys
from pathlib import Path

import pytest

from nanasqlite import CacheType, NanaSQLite

REPO_ROOT = Path(__file__).parent.parent


class TestNanaSQLiteInitValidatorDefault:
    """NanaSQLite.__init__ の validator デフォルトが None であることを確認する。"""

    def test_validator_default_is_none(self):
        """NanaSQLite.__init__ の validator パラメーターのデフォルトが None であること。

        _UNSET は table() での「省略 = 親継承」セマンティクスのためのセンチネルであり、
        コンストラクターでは親から継承する概念がないため、None をデフォルトにすべき。
        """
        sig = inspect.signature(NanaSQLite.__init__)
        validator_param = sig.parameters.get("validator")
        assert validator_param is not None, "NanaSQLite.__init__ に validator パラメーターが存在しない"
        default = validator_param.default
        assert default is None, (
            f"NanaSQLite.__init__ の validator デフォルトが None ではなく {default!r} になっている。"
            " _UNSET は table() 専用のセンチネルであり、コンストラクターでは None を使用すべき。"
        )

    def test_no_validator_means_no_validation(self, tmp_path):
        """引数なしで NanaSQLite を生成すると _validator が None になること。"""
        db = NanaSQLite(str(tmp_path / "test.db"), table="t")
        assert db._validator is None

    def test_explicit_none_validator_works(self, tmp_path):
        """validator=None を明示的に渡すと _validator が None になること。"""
        db = NanaSQLite(str(tmp_path / "test.db"), table="t", validator=None)
        assert db._validator is None


class TestTableCacheTTLParam:
    """table() に cache_ttl / cache_persistence_ttl パラメーターがあることを確認する。"""

    def test_table_signature_has_cache_ttl(self):
        """table() に cache_ttl パラメーターが存在すること。"""
        sig = inspect.signature(NanaSQLite.table)
        assert "cache_ttl" in sig.parameters, (
            "NanaSQLite.table() に cache_ttl パラメーターが存在しない。"
            " 親が非TTLの場合でも cache_strategy=TTL で table() を呼べるよう追加してください。"
        )

    def test_table_signature_has_cache_persistence_ttl(self):
        """table() に cache_persistence_ttl パラメーターが存在すること。"""
        sig = inspect.signature(NanaSQLite.table)
        assert "cache_persistence_ttl" in sig.parameters, (
            "NanaSQLite.table() に cache_persistence_ttl パラメーターが存在しない。"
        )

    def test_table_can_set_ttl_strategy_on_non_ttl_parent(self, tmp_path):
        """非TTL親から table() で TTL 戦略の子を作れること。"""
        parent = NanaSQLite(str(tmp_path / "test.db"), table="parent")
        # 親は UNBOUNDED (デフォルト)。子は TTL 戦略を明示的に指定 — ValueError が出ないこと
        child = parent.table("child", cache_strategy=CacheType.TTL, cache_ttl=30.0)
        assert child._cache_ttl_raw == 30.0

    def test_table_ttl_strategy_without_cache_ttl_raises(self, tmp_path):
        """非TTL親から cache_strategy=TTL だけ指定して table() を呼ぶと ValueError が発生すること。"""
        parent = NanaSQLite(str(tmp_path / "test.db"), table="parent")
        with pytest.raises(ValueError, match="cache_ttl"):
            parent.table("child", cache_strategy=CacheType.TTL)

    def test_table_inherits_parent_ttl(self, tmp_path):
        """TTL 親から省略で table() を呼ぶと cache_ttl が継承されること。"""
        parent = NanaSQLite(
            str(tmp_path / "test.db"), table="parent",
            cache_strategy=CacheType.TTL, cache_ttl=60.0,
        )
        child = parent.table("child")
        assert child._cache_ttl_raw == 60.0

    def test_table_inherits_parent_persistence_ttl(self, tmp_path):
        """TTL 親の cache_persistence_ttl も子に継承されること。"""
        parent = NanaSQLite(
            str(tmp_path / "test.db"), table="parent",
            cache_strategy=CacheType.TTL, cache_ttl=60.0, cache_persistence_ttl=True,
        )
        child = parent.table("child")
        assert child._cache_persistence_ttl_raw is True

    def test_table_explicit_cache_ttl_overrides_parent(self, tmp_path):
        """TTL 親から table() で cache_ttl を上書きできること。"""
        parent = NanaSQLite(
            str(tmp_path / "test.db"), table="parent",
            cache_strategy=CacheType.TTL, cache_ttl=60.0,
        )
        child = parent.table("child", cache_ttl=10.0)
        assert child._cache_ttl_raw == 10.0


class TestNoUnusedImports:
    """Ruff F401: 未使用インポートが存在しないことを確認する。"""

    def _run_ruff_f401(self, filepath: str) -> str:
        """指定ファイルに対して ruff F401 チェックを実行し、エラーがあれば返す。

        ruff が利用できない環境（devextras 未インストール）ではテストをスキップする。
        """
        if shutil.which("ruff") is None:
            pytest.skip("ruff is not installed — skipping F401 check")
        result = subprocess.run(
            [sys.executable, "-m", "ruff", "check", filepath, "--select", "F401"],
            capture_output=True,
            text=True,
            cwd=str(REPO_ROOT),
        )
        return result.stdout if result.returncode != 0 else ""

    def test_no_unused_import_time_in_comprehensive(self):
        """test_table_inheritance_comprehensive.py に未使用の `import time` がないこと。"""
        errors = self._run_ruff_f401("tests/test_table_inheritance_comprehensive.py")
        assert not errors, f"Ruff F401 errors:\n{errors}"

    def test_no_unused_import_pytest_in_tdd_fixes(self):
        """test_tdd_review_fixes.py に未使用の `import pytest` がないこと。"""
        errors = self._run_ruff_f401("tests/test_tdd_review_fixes.py")
        assert not errors, f"Ruff F401 errors:\n{errors}"



"""
table() 継承セマンティクスとキャッシュ戦略の包括的テスト
(Comprehensive tests for table() inheritance semantics and cache strategies)

このテストファイルは以下を網羅的にテストします:
1. table() のキャッシュ戦略継承 (UNBOUNDED / LRU / TTL)
   - TTL: cache_ttl / cache_persistence_ttl の継承（省略時の ValueError 防止）
2. table() の validator / coerce 継承セマンティクス
3. table() の cache_strategy 型アノテーション（"ttl" 文字列含む）
4. HAS_VALIDKIT / HAS_ORJSON フラグの整合性
5. optional import の例外絞り込み（ImportError のみ捕捉）
6. per-table バリデーション全パス
7. coerce 動作の全パス（デュアル要件）
8. batch_update() バリデーション（アトミック失敗）
9. AsyncNanaSQLite の各種パス
"""
from __future__ import annotations

import pytest

from nanasqlite import CacheType, NanaSQLite

# validkit availability detection (try/except import, not find_spec)
try:
    import validkit  # noqa: F401

    validkit_installed = True
except ImportError:
    validkit_installed = False


# ===========================================================================
# セクション 1: table() キャッシュ戦略継承
# ===========================================================================


class TestTableCacheStrategyInheritance:
    """table() が親のキャッシュ戦略を正しく継承することを確認する。"""

    def test_table_inherits_unbounded_strategy(self, tmp_path):
        """親が UNBOUNDED の場合、子も UNBOUNDED を継承する。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.UNBOUNDED)
        child = db.table("child")
        assert child._cache_strategy_raw == CacheType.UNBOUNDED
        # 子テーブルに書き込みと読み込みが正常に動作する
        child["k"] = "v"
        assert child["k"] == "v"

    def test_table_inherits_lru_strategy(self, tmp_path):
        """親が LRU の場合、子も LRU を継承し同じサイズを使う。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=10)
        child = db.table("child")
        assert child._cache_strategy_raw == CacheType.LRU
        assert child._cache_size_raw == 10
        child["a"] = 1
        child["b"] = 2
        assert child["a"] == 1

    def test_table_override_cache_strategy(self, tmp_path):
        """table() で明示的に cache_strategy を渡すと、親の戦略を上書きできる。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.UNBOUNDED)
        child = db.table("child", cache_strategy=CacheType.LRU, cache_size=5)
        assert child._cache_strategy_raw == CacheType.LRU
        assert child._cache_size_raw == 5

    def test_table_inherits_lru_eviction(self, tmp_path):
        """LRU 継承した子テーブルでエビクションが正しく機能する。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=3)
        child = db.table("child")
        for i in range(5):
            child[f"k{i}"] = i
        # DB にはすべての値が保存されている
        for i in range(5):
            assert child[f"k{i}"] == i

    def test_table_override_with_lru_string(self, tmp_path):
        """table() に文字列 'lru' を渡して LRU 戦略を指定できる。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        child = db.table("child", cache_strategy="lru", cache_size=5)
        child["x"] = 42
        assert child["x"] == 42

    def test_table_override_with_unbounded_string(self, tmp_path):
        """table() に文字列 'unbounded' を渡して UNBOUNDED 戦略を指定できる。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=5)
        child = db.table("child", cache_strategy="unbounded")
        child["x"] = "hello"
        assert child["x"] == "hello"


class TestTableTTLCacheInheritance:
    """table() が TTL キャッシュ設定を正しく継承することを確認する（TTL ValueError 防止）。"""

    def test_table_inherits_ttl_strategy_does_not_raise(self, tmp_path):
        """親が TTL 戦略の場合、table() 呼び出しが ValueError を送出しないこと。

        これは修正されたバグ: cache_ttl を継承しないと ValueError が発生していた。
        """
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
        )
        # table() は ValueError を送出してはいけない
        child = db.table("child")
        assert child._cache_strategy_raw == CacheType.TTL
        assert child._cache_ttl_raw == 60.0

    def test_table_inherits_ttl_value(self, tmp_path):
        """子テーブルが親の cache_ttl 値を正しく継承する。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=30.0,
        )
        child = db.table("child")
        assert child._cache_ttl_raw == 30.0

    def test_table_inherits_ttl_with_persistence(self, tmp_path):
        """子テーブルが cache_persistence_ttl も継承する。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
            cache_persistence_ttl=True,
        )
        child = db.table("child")
        assert child._cache_strategy_raw == CacheType.TTL
        assert child._cache_ttl_raw == 60.0
        assert child._cache_persistence_ttl_raw is True

    def test_table_ttl_child_can_read_write(self, tmp_path):
        """TTL を継承した子テーブルで読み書きが正常に動作する。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
        )
        child = db.table("child")
        child["key"] = {"value": 123}
        assert child["key"] == {"value": 123}

    def test_table_override_ttl_strategy(self, tmp_path):
        """table() で TTL を上書き指定する場合、親が TTL なら cache_ttl も継承される。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=30.0,
        )
        # 子で TTL を明示し、cache_ttl は親から継承される（ValueError にならない）
        child = db.table("child", cache_strategy=CacheType.TTL)
        assert child._cache_strategy_raw == CacheType.TTL
        assert child._cache_ttl_raw == 30.0

    def test_table_inherits_ttl_string_strategy(self, tmp_path):
        """文字列 'ttl' で指定された親の戦略も table() で継承される。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy="ttl",
            cache_ttl=20.0,
        )
        child = db.table("child")
        assert child._cache_strategy_raw == "ttl"
        assert child._cache_ttl_raw == 20.0
        child["z"] = 99
        assert child["z"] == 99

    def test_multiple_children_from_ttl_parent(self, tmp_path):
        """TTL 親から複数の子テーブルを作成しても全て成功する。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
        )
        children = [db.table(f"child{i}") for i in range(5)]
        for i, child in enumerate(children):
            child[f"k{i}"] = i * 10
            assert child[f"k{i}"] == i * 10


class TestTableCacheSizeInheritance:
    """table() が親の cache_size を正しく継承することを確認する。"""

    def test_table_inherits_cache_size(self, tmp_path):
        """子テーブルが親の cache_size を継承する。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=20)
        child = db.table("child")
        assert child._cache_size_raw == 20

    def test_table_overrides_cache_size(self, tmp_path):
        """table() で cache_size を明示指定すると親の値を上書きできる。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=20)
        child = db.table("child", cache_size=5)
        assert child._cache_size_raw == 5

    def test_table_none_cache_size_inherits_parent(self, tmp_path):
        """cache_size=None を渡さずに省略すると、親の値が継承される。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=15)
        child = db.table("child")
        assert child._cache_size_raw == 15


# ===========================================================================
# セクション 2: _UNSET センチネル / validator・coerce 継承
# ===========================================================================


class TestTableValidatorInheritance:
    """table() での validator / coerce の継承と上書きを確認する。"""

    def test_table_inherits_none_validator(self, tmp_path):
        """親に validator がない場合、子も validator なしで動作する。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        child = db.table("child")
        assert child._validator is None
        child["k"] = {"any": "value"}
        assert child["k"]["any"] == "value"

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    def test_table_inherits_validator_from_parent(self, tmp_path):
        """親の validator が子テーブルに継承される。"""
        from validkit import v

        schema = {"name": v.str()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        child = db.table("child")
        assert child._validator is schema

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    def test_table_explicit_none_overrides_validator(self, tmp_path):
        """table(validator=None) は親の validator を無効化する。"""
        from validkit import v

        schema = {"name": v.str()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        child = db.table("child", validator=None)
        assert child._validator is None
        # バリデーションなしで任意の値が書き込める
        child["k"] = {"anything": 42}
        assert child["k"]["anything"] == 42

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    def test_table_override_with_new_validator(self, tmp_path):
        """table() で別のスキーマを指定すると、子テーブルはそのスキーマを使う。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        parent_schema = {"x": v.int()}
        child_schema = {"name": v.str()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=parent_schema)
        child = db.table("child", validator=child_schema)

        child["c"] = {"name": "Alice"}
        assert child["c"] == {"name": "Alice"}

        with pytest.raises(NanaSQLiteValidationError):
            child["bad"] = {"name": 999}

    def test_table_inherits_coerce_false(self, tmp_path):
        """親の coerce=False (デフォルト) が子に継承される。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        child = db.table("child")
        assert child._coerce is False

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    def test_table_inherits_coerce_true(self, tmp_path):
        """親の coerce=True が子に継承される。"""
        from validkit import v

        schema = {"age": v.int().coerce()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        child = db.table("child")
        assert child._coerce is True
        child["c"] = {"age": "5"}
        assert child["c"] == {"age": 5}

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    def test_table_override_coerce_to_false(self, tmp_path):
        """table(coerce=False) で子テーブルの coerce を無効化できる。"""
        from validkit import v

        schema = {"age": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        child = db.table("child", coerce=False)
        assert child._coerce is False
        # coerce=False なので正しい型のみ受け付ける
        child["c"] = {"age": 10}
        assert child["c"] == {"age": 10}


# ===========================================================================
# セクション 3: HAS_VALIDKIT / HAS_ORJSON フラグ整合性
# ===========================================================================


class TestOptionalImportFlags:
    """HAS_VALIDKIT / HAS_ORJSON フラグが正しく設定されていることを確認する。"""

    def test_has_validkit_flag_is_bool(self):
        """HAS_VALIDKIT は bool 型であること。"""
        from nanasqlite import core as core_mod

        assert isinstance(core_mod.HAS_VALIDKIT, bool)

    def test_has_validkit_matches_actual_import(self):
        """HAS_VALIDKIT の値が実際の validkit インポート成否と一致する。"""
        from nanasqlite import HAS_VALIDKIT

        assert HAS_VALIDKIT is validkit_installed

    def test_has_orjson_flag_is_bool(self):
        """HAS_ORJSON は bool 型であること。"""
        from nanasqlite import core as core_mod

        assert isinstance(core_mod.HAS_ORJSON, bool)

    def test_has_orjson_matches_actual_import(self):
        """HAS_ORJSON の値が実際の orjson インポート成否と一致する。"""
        from nanasqlite import core as core_mod

        try:
            import orjson  # noqa: F401
            expected = True
        except ImportError:
            expected = False
        assert core_mod.HAS_ORJSON is expected

    def test_has_validkit_exportable_from_package(self):
        """HAS_VALIDKIT は nanasqlite パッケージから直接インポートできる。"""
        from nanasqlite import HAS_VALIDKIT  # noqa: F401 — must not raise

    @pytest.mark.skipif(validkit_installed, reason="validkit-py がインストールされている環境ではスキップ")
    def test_import_error_raised_when_validkit_missing(self, tmp_path):
        """HAS_VALIDKIT=False 環境で validator を渡すと ImportError が送出される。"""
        with pytest.raises(ImportError, match="validkit-py"):
            NanaSQLite(str(tmp_path / "err.db"), validator={"key": "val"})

    @pytest.mark.skipif(validkit_installed, reason="validkit-py がインストールされている環境ではスキップ")
    @pytest.mark.asyncio
    async def test_async_import_error_raised_when_validkit_missing(self, tmp_path):
        """HAS_VALIDKIT=False 環境で AsyncNanaSQLite に validator を渡すと ImportError が送出される。"""
        from nanasqlite import AsyncNanaSQLite

        with pytest.raises(ImportError, match="validkit-py"):
            AsyncNanaSQLite(str(tmp_path / "async_err.db"), validator={"key": "val"})


# ===========================================================================
# セクション 4: _cache_strategy_raw の保持と継承チェーン
# ===========================================================================


class TestCacheRawAttributeStorage:
    """__init__ で _cache_*_raw 属性が正しく保存されることを確認する。"""

    def test_cache_strategy_raw_stored_unbounded(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"))
        assert db._cache_strategy_raw == CacheType.UNBOUNDED

    def test_cache_strategy_raw_stored_lru(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=5)
        assert db._cache_strategy_raw == CacheType.LRU

    def test_cache_strategy_raw_stored_ttl(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.TTL, cache_ttl=30.0)
        assert db._cache_strategy_raw == CacheType.TTL

    def test_cache_size_raw_stored(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=7)
        assert db._cache_size_raw == 7

    def test_cache_size_raw_none_when_not_provided(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"))
        assert db._cache_size_raw is None

    def test_cache_ttl_raw_stored(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.TTL, cache_ttl=45.0)
        assert db._cache_ttl_raw == 45.0

    def test_cache_ttl_raw_none_when_not_provided(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"))
        assert db._cache_ttl_raw is None

    def test_cache_persistence_ttl_raw_stored(self, tmp_path):
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
            cache_persistence_ttl=True,
        )
        assert db._cache_persistence_ttl_raw is True

    def test_cache_persistence_ttl_raw_default_false(self, tmp_path):
        db = NanaSQLite(str(tmp_path / "db.db"))
        assert db._cache_persistence_ttl_raw is False


# ===========================================================================
# セクション 5: バリデーション全パス (validkit 必須)
# ===========================================================================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
class TestValidkitValidationPaths:
    """validkit-py バリデーションの全パスを網羅するテスト。"""

    def test_valid_value_is_accepted(self, tmp_path):
        """スキーマに一致する値は正常に保存される。"""
        from validkit import v

        schema = {"name": v.str(), "age": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        db["u"] = {"name": "Alice", "age": 30}
        assert db["u"] == {"name": "Alice", "age": 30}

    def test_invalid_value_raises_validation_error(self, tmp_path):
        """スキーマ違反の値は NanaSQLiteValidationError を送出する。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        schema = {"name": v.str(), "age": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        with pytest.raises(NanaSQLiteValidationError):
            db["u"] = {"name": "Bob", "age": "bad"}

    def test_invalid_value_not_written_to_db(self, tmp_path):
        """バリデーションエラー時は DB に値が保存されない。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        schema = {"score": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        with pytest.raises(NanaSQLiteValidationError):
            db["e"] = {"score": "oops"}
        assert "e" not in db

    def test_validator_none_explicit_no_validation(self, tmp_path):
        """validator=None を明示すると任意の値が保存される。"""
        db = NanaSQLite(str(tmp_path / "db.db"), validator=None)
        db["k"] = {"anything": [1, 2, 3]}
        assert db["k"]["anything"] == [1, 2, 3]

    def test_validator_not_provided_no_validation(self, tmp_path):
        """validator を省略した場合も任意の値が保存される。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        db["k"] = {"x": True, "y": "z"}
        assert db["k"]["x"] is True

    def test_multiple_schemas_independent(self, tmp_path):
        """異なるスキーマを持つ複数のインスタンスが互いに独立している。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        schema_a = {"x": v.int()}
        schema_b = {"name": v.str()}
        db_a = NanaSQLite(str(tmp_path / "a.db"), validator=schema_a)
        db_b = NanaSQLite(str(tmp_path / "b.db"), validator=schema_b)

        db_a["k"] = {"x": 1}
        db_b["k"] = {"name": "hello"}

        with pytest.raises(NanaSQLiteValidationError):
            db_a["bad"] = {"x": "not_int"}
        with pytest.raises(NanaSQLiteValidationError):
            db_b["bad"] = {"name": 999}

    def test_batch_update_atomic_failure(self, tmp_path):
        """batch_update() は1件でもバリデーション違反があれば全件書き込まない。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        schema = {"age": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        with pytest.raises(NanaSQLiteValidationError):
            db.batch_update({
                "u1": {"age": 20},
                "u2": {"age": "bad"},
            })
        assert "u1" not in db
        assert "u2" not in db

    def test_batch_update_all_valid_writes_all(self, tmp_path):
        """batch_update() は全件バリデーションが通れば全件書き込む。"""
        from validkit import v

        schema = {"age": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema)
        db.batch_update({
            "u1": {"age": 20},
            "u2": {"age": 30},
        })
        assert db["u1"]["age"] == 20
        assert db["u2"]["age"] == 30

    def test_per_table_validator_isolation(self, tmp_path):
        """異なるテーブルで別スキーマを使うと、テーブルごとにバリデーションが独立する。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        parent_schema = {"x": v.int()}
        child_schema = {"name": v.str()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=parent_schema)
        child = db.table("child", validator=child_schema)

        db["p"] = {"x": 1}
        child["c"] = {"name": "hello"}

        with pytest.raises(NanaSQLiteValidationError):
            db["bad"] = {"x": "not_int"}
        with pytest.raises(NanaSQLiteValidationError):
            child["bad"] = {"name": 999}


# ===========================================================================
# セクション 6: coerce 全パス (dual-requirement)
# ===========================================================================


@pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
class TestCoercePaths:
    """coerce モードの全パスを網羅するテスト。"""

    def test_coerce_true_with_field_coerce_converts(self, tmp_path):
        """NanaSQLite coerce=True + フィールド .coerce() で型変換が有効になる。"""
        from validkit import v

        schema = {"age": v.int().coerce()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        db["u"] = {"age": "42"}
        assert db["u"] == {"age": 42}

    def test_coerce_true_without_field_coerce_fails(self, tmp_path):
        """NanaSQLite coerce=True でもフィールドに .coerce() がなければ型不一致は失敗する。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        schema = {"age": v.int()}  # .coerce() なし
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        with pytest.raises(NanaSQLiteValidationError):
            db["u"] = {"age": "42"}
        # 正しい型は通る
        db["u"] = {"age": 42}
        assert db["u"] == {"age": 42}

    def test_coerce_false_with_field_coerce_stores_original(self, tmp_path):
        """coerce=False (デフォルト) + フィールド .coerce() の場合、変換後ではなく元の値が保存される。"""
        from validkit import v

        schema = {"age": v.int().coerce()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=False)
        db["u"] = {"age": "30"}
        assert db["u"] == {"age": "30"}  # 元の文字列が保存される

    def test_coerce_inherited_in_batch_update(self, tmp_path):
        """coerce=True は batch_update() でも機能する。"""
        from validkit import v

        schema = {"score": v.float().coerce()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        db.batch_update({"a": {"score": "1.5"}, "b": {"score": "2.0"}})
        assert db["a"] == {"score": 1.5}
        assert db["b"] == {"score": 2.0}

    def test_coerce_inherited_by_child_table(self, tmp_path):
        """子テーブルは親の coerce=True を継承する。"""
        from validkit import v

        schema = {"age": v.int().coerce()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        child = db.table("child")
        child["c"] = {"age": "7"}
        assert child["c"] == {"age": 7}

    def test_coerce_can_be_disabled_in_child(self, tmp_path):
        """table(coerce=False) で子テーブルの coerce を無効化できる。"""
        from validkit import v

        schema = {"age": v.int()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=True)
        child = db.table("child", coerce=False)
        assert child._coerce is False
        child["c"] = {"age": 5}
        assert child["c"] == {"age": 5}

    def test_coerce_can_be_enabled_in_child(self, tmp_path):
        """table(coerce=True) で子テーブルの coerce を有効化できる。"""
        from validkit import v

        schema = {"age": v.int().coerce()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=schema, coerce=False)
        child = db.table("child", coerce=True)
        assert child._coerce is True
        child["c"] = {"age": "8"}
        assert child["c"] == {"age": 8}


# ===========================================================================
# セクション 7: AsyncNanaSQLite 継承パス
# ===========================================================================


class TestAsyncTableInheritance:
    """AsyncNanaSQLite の table() 継承セマンティクスを確認する。"""

    @pytest.mark.asyncio
    async def test_async_table_inherits_cache_strategy(self, tmp_path):
        """AsyncNanaSQLite: table() が親のキャッシュ戦略を継承する。"""
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.LRU,
            cache_size=10,
        ) as db:
            child = await db.table("child")
            await child.aset("k", "v")
            assert await child.aget("k") == "v"

    @pytest.mark.asyncio
    async def test_async_table_inherits_ttl_does_not_raise(self, tmp_path):
        """AsyncNanaSQLite: TTL 戦略の親から table() を作成しても ValueError が出ない。"""
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
        ) as db:
            child = await db.table("child")
            await child.aset("key", "value")
            assert await child.aget("key") == "value"

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    @pytest.mark.asyncio
    async def test_async_table_inherits_validator(self, tmp_path):
        """AsyncNanaSQLite: 親の validator が子テーブルに継承される。"""
        from validkit import v

        from nanasqlite import AsyncNanaSQLite, NanaSQLiteValidationError

        schema = {"name": v.str()}
        async with AsyncNanaSQLite(str(tmp_path / "db.db"), validator=schema) as db:
            child = await db.table("child")
            await child.aset("c", {"name": "hello"})
            assert (await child.aget("c"))["name"] == "hello"
            with pytest.raises(NanaSQLiteValidationError):
                await child.aset("bad", {"name": 999})

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    @pytest.mark.asyncio
    async def test_async_table_explicit_none_disables_validator(self, tmp_path):
        """AsyncNanaSQLite: table(validator=None) で親の validator を無効化できる。"""
        from validkit import v

        from nanasqlite import AsyncNanaSQLite

        schema = {"name": v.str()}
        async with AsyncNanaSQLite(str(tmp_path / "db.db"), validator=schema) as db:
            child = await db.table("child", validator=None)
            await child.aset("k", {"anything": 42})
            assert (await child.aget("k"))["anything"] == 42

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    @pytest.mark.asyncio
    async def test_async_coerce_inherited(self, tmp_path):
        """AsyncNanaSQLite: coerce=True が子テーブルに継承される。"""
        from validkit import v

        from nanasqlite import AsyncNanaSQLite

        schema = {"age": v.int().coerce()}
        async with AsyncNanaSQLite(
            str(tmp_path / "db.db"), validator=schema, coerce=True
        ) as db:
            child = await db.table("child")
            await child.aset("c", {"age": "12"})
            assert (await child.aget("c")) == {"age": 12}

    @pytest.mark.skipif(validkit_installed, reason="validkit-py がインストールされている環境ではスキップ")
    @pytest.mark.asyncio
    async def test_async_import_error_on_constructor_when_validkit_missing(self, tmp_path):
        """validkit-py が未インストールの場合、AsyncNanaSQLite コンストラクタで ImportError が出る。"""
        from nanasqlite import AsyncNanaSQLite

        with pytest.raises(ImportError, match="validkit-py"):
            AsyncNanaSQLite(str(tmp_path / "err.db"), validator={"k": "v"})


# ===========================================================================
# セクション 8: table() の cache_strategy 型アノテーション（文字列値）
# ===========================================================================


class TestTableCacheStrategyStringAnnotation:
    """table() の cache_strategy に文字列 'ttl' / 'lru' / 'unbounded' を渡せることを確認する。"""

    def test_table_accepts_ttl_string(self, tmp_path):
        """table() に 'ttl' 文字列を渡せる（ValueError が出ない）。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.TTL, cache_ttl=60.0)
        # TTL を親から継承した子を 'ttl' 文字列でも作れる
        child = db.table("child", cache_strategy="ttl")
        # cache_ttl は親から継承されるので ValueError にならない
        child["k"] = "v"
        assert child["k"] == "v"

    def test_table_accepts_lru_string(self, tmp_path):
        """table() に 'lru' 文字列と cache_size を渡せる。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        child = db.table("child", cache_strategy="lru", cache_size=10)
        child["k"] = "v"
        assert child["k"] == "v"

    def test_table_accepts_unbounded_string(self, tmp_path):
        """table() に 'unbounded' 文字列を渡せる。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        child = db.table("child", cache_strategy="unbounded")
        child["k"] = "v"
        assert child["k"] == "v"


# ===========================================================================
# セクション 9: 接続共有と多テーブル同時操作
# ===========================================================================


class TestMultiTableOperations:
    """同一 DB から複数テーブルを操作する際の動作を確認する。"""

    def test_multiple_tables_share_connection(self, tmp_path):
        """複数の子テーブルが同じ DB ファイルを共有して動作する。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        users = db.table("users")
        products = db.table("products")
        users["u1"] = {"name": "Alice"}
        products["p1"] = {"name": "Laptop"}
        assert users["u1"]["name"] == "Alice"
        assert products["p1"]["name"] == "Laptop"

    def test_tables_data_are_isolated(self, tmp_path):
        """異なるテーブルのデータは独立していて、相互に影響しない。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        t1 = db.table("t1")
        t2 = db.table("t2")
        t1["k"] = "table1_value"
        t2["k"] = "table2_value"
        assert t1["k"] == "table1_value"
        assert t2["k"] == "table2_value"

    def test_parent_and_child_tables_independent_data(self, tmp_path):
        """親テーブルと子テーブルのデータは独立している。"""
        db = NanaSQLite(str(tmp_path / "db.db"))
        child = db.table("child")
        db["shared_key"] = "parent_value"
        child["shared_key"] = "child_value"
        assert db["shared_key"] == "parent_value"
        assert child["shared_key"] == "child_value"

    @pytest.mark.skipif(not validkit_installed, reason="validkit-py が未インストールのためスキップ")
    def test_child_schema_does_not_affect_parent(self, tmp_path):
        """子テーブルのスキーマを変更しても親インスタンスのバリデーションは変わらない。"""
        from validkit import v

        from nanasqlite import NanaSQLiteValidationError

        parent_schema = {"x": v.int()}
        child_schema = {"name": v.str()}
        db = NanaSQLite(str(tmp_path / "db.db"), validator=parent_schema)
        _ = db.table("child", validator=child_schema)

        db["p"] = {"x": 42}
        with pytest.raises(NanaSQLiteValidationError):
            db["bad"] = {"x": "not_int"}

    def test_deep_chain_table_inherits_strategy(self, tmp_path):
        """table() の連鎖（孫テーブル）でもキャッシュ戦略が正しく設定される。"""
        db = NanaSQLite(str(tmp_path / "db.db"), cache_strategy=CacheType.LRU, cache_size=10)
        child = db.table("child")
        # 孫テーブルも同じ戦略を持つ
        grandchild = child.table("grandchild")
        assert grandchild._cache_strategy_raw == CacheType.LRU
        grandchild["k"] = "v"
        assert grandchild["k"] == "v"

    def test_deep_chain_ttl_does_not_raise(self, tmp_path):
        """table() の連鎖（孫テーブル）でも TTL が正しく継承されて ValueError が出ない。"""
        db = NanaSQLite(
            str(tmp_path / "db.db"),
            cache_strategy=CacheType.TTL,
            cache_ttl=60.0,
        )
        child = db.table("child")
        grandchild = child.table("grandchild")
        assert grandchild._cache_ttl_raw == 60.0
        grandchild["k"] = "v"
        assert grandchild["k"] == "v"


# ===========================================================================
# セクション 10: AsyncNanaSQLite 基本動作 (バリデーションなし)
# ===========================================================================


class TestAsyncBasicOperations:
    """AsyncNanaSQLite の基本的な読み書き操作を確認する。"""

    @pytest.mark.asyncio
    async def test_async_set_and_get(self, tmp_path):
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(str(tmp_path / "db.db")) as db:
            await db.aset("k", {"val": 1})
            result = await db.aget("k")
            assert result == {"val": 1}

    @pytest.mark.asyncio
    async def test_async_contains(self, tmp_path):
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(str(tmp_path / "db.db")) as db:
            assert not await db.acontains("k")
            await db.aset("k", 42)
            assert await db.acontains("k")

    @pytest.mark.asyncio
    async def test_async_delete(self, tmp_path):
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(str(tmp_path / "db.db")) as db:
            await db.aset("k", 99)
            await db.adelete("k")
            assert not await db.acontains("k")

    @pytest.mark.asyncio
    async def test_async_batch_update(self, tmp_path):
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(str(tmp_path / "db.db")) as db:
            await db.aupdate({"a": 1, "b": 2})
            assert await db.aget("a") == 1
            assert await db.aget("b") == 2

    @pytest.mark.asyncio
    async def test_async_len(self, tmp_path):
        from nanasqlite import AsyncNanaSQLite

        async with AsyncNanaSQLite(str(tmp_path / "db.db")) as db:
            await db.aset("a", 1)
            await db.aset("b", 2)
            assert await db.alen() == 2



"""TDD cycle 8 — Fix CI failures from test_tdd_cycle_7.py.

Issues found in CI:
1. Ruff I001: import block in test_tdd_cycle_7.py was unsorted (now fixed).
2. test_mypy_installed hard-asserts mypy is installed → fails in CI where mypy
   is not a test-job dependency (only a dev/lint dependency).

Fix:
1. ruff --fix applied (import sort).
2. test_mypy_installed must call _skip_if_mypy_not_installed() so it skips
   gracefully when mypy is absent, rather than failing with AssertionError.

TDD: all tests written BEFORE the fix is applied.
"""

import ast
import shutil
import subprocess
import sys
from pathlib import Path

import pytest

REPO_ROOT = Path(__file__).parent.parent
TEST_CYCLE_7 = REPO_ROOT / "tests" / "test_tdd_cycle_7.py"


class TestRuffCleanCycle7:
    """tests/test_tdd_cycle_7.py must pass ruff check (I001 fix)."""

    def test_ruff_clean(self):
        """ruff check must report no errors on test_tdd_cycle_7.py."""
        if shutil.which("ruff") is None:
            pytest.skip("ruff not installed")
        result = subprocess.run(
            [sys.executable, "-m", "ruff", "check", str(TEST_CYCLE_7)],
            cwd=str(REPO_ROOT),
            capture_output=True,
            text=True,
        )
        assert result.returncode == 0, (
            f"ruff errors in test_tdd_cycle_7.py:\n{result.stdout}"
        )


class TestMypyInstalledSkipsGracefully:
    """test_mypy_installed must skip (not fail) when mypy is absent."""

    def _get_test_mypy_installed_node(self) -> ast.FunctionDef:
        """Parse test_tdd_cycle_7.py and return the test_mypy_installed function node."""
        source = TEST_CYCLE_7.read_text(encoding="utf-8")
        tree = ast.parse(source)
        for node in ast.walk(tree):
            if (
                isinstance(node, ast.ClassDef)
                and node.name == "TestMypyPassesWithEllipsisType"
            ):
                for item in node.body:
                    if (
                        isinstance(item, ast.FunctionDef)
                        and item.name == "test_mypy_installed"
                    ):
                        return item
        raise AssertionError(
            "test_mypy_installed not found in TestMypyPassesWithEllipsisType"
        )

    def test_mypy_installed_calls_skip_helper(self):
        """test_mypy_installed must call _skip_if_mypy_not_installed() so CI skips it."""
        func_node = self._get_test_mypy_installed_node()
        # Look for a call to _skip_if_mypy_not_installed in the function body
        skip_calls = [
            n
            for n in ast.walk(func_node)
            if isinstance(n, ast.Call)
            and isinstance(n.func, ast.Name)
            and n.func.id == "_skip_if_mypy_not_installed"
        ]
        assert len(skip_calls) > 0, (
            "test_mypy_installed does not call _skip_if_mypy_not_installed(). "
            "Add _skip_if_mypy_not_installed() at the start of the function so it "
            "skips gracefully when mypy is absent in CI."
        )

    def test_mypy_installed_has_no_unconditional_hard_assert(self):
        """test_mypy_installed must not unconditionally assert mypy spec is not None."""
        func_node = self._get_test_mypy_installed_node()
        # Collect assert statements that check find_spec("mypy") is not None
        # before any skip call
        first_stmt = func_node.body[0] if func_node.body else None
        if first_stmt is None:
            return
        # If the very first statement is an unconditional assert on mypy's find_spec,
        # that is the pattern that fails in CI.
        if isinstance(first_stmt, ast.Assert):
            # Check if the test value is a comparison involving find_spec
            test_src = ast.unparse(first_stmt.test)
            assert "find_spec" not in test_src or "_skip_if" in "".join(
                ast.unparse(n) for n in func_node.body
            ), (
                "test_mypy_installed starts with an unconditional assert on "
                "importlib.util.find_spec('mypy'), which fails in CI when mypy is "
                "not installed. Use _skip_if_mypy_not_installed() instead."
            )



"""
v1.3.4b1 新機能テスト

- P2-1: lock_timeout パラメータ
- P2-3: backup / restore メソッド
"""

import os
import threading

import pytest

from nanasqlite import NanaSQLite, NanaSQLiteLockError
from nanasqlite.exceptions import (
    NanaSQLiteClosedError,
    NanaSQLiteConnectionError,
    NanaSQLiteDatabaseError,
    NanaSQLiteTransactionError,
    NanaSQLiteValidationError,
)

# ===========================================================
# P2-1: lock_timeout パラメータ
# ===========================================================


def test_lock_timeout_default_none(tmp_path):
    """lock_timeout デフォルト (None) で通常操作が問題なく動作すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"
    assert db["key"] == "value"
    db.close()


def test_lock_timeout_positive(tmp_path):
    """lock_timeout に正の値を設定して通常操作が問題なく動作すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=5.0)
    db["key"] = "value"
    assert db["key"] == "value"
    db.close()


def test_lock_timeout_raises_on_deadlock(tmp_path):
    """ロックを保持したまま別スレッドが lock_timeout 内に取得できない場合 NanaSQLiteLockError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.2)
    errors = []

    lock_held = threading.Event()   # hold_lock がロックを取得したことを通知
    can_release = threading.Event()  # try_acquire が完了したことを通知（リリース許可）

    def hold_lock():
        db._lock.acquire()
        try:
            lock_held.set()           # ロック取得を通知
            can_release.wait()        # try_acquire が終わるまで保持
        finally:
            db._lock.release()

    def try_acquire():
        lock_held.wait()          # hold_lock がロックを取得するまで待つ
        try:
            with db._acquire_lock():
                pass
        except NanaSQLiteLockError as e:
            errors.append(e)
        finally:
            can_release.set()     # hold_lock にリリースを許可

    t1 = threading.Thread(target=hold_lock)
    t2 = threading.Thread(target=try_acquire)
    t1.start()
    t2.start()
    t2.join(timeout=5.0)
    assert not t2.is_alive(), "try_acquire thread did not finish in time"
    t1.join(timeout=5.0)
    assert not t1.is_alive(), "hold_lock thread did not finish in time"

    assert len(errors) == 1
    assert "0.2s" in str(errors[0])
    db.close()


def test_lock_timeout_negative_raises(tmp_path):
    """lock_timeout に負値を渡した場合 NanaSQLiteValidationError が発生すること"""
    with pytest.raises(NanaSQLiteValidationError):
        NanaSQLite(str(tmp_path / "test.db"), lock_timeout=-1.0)


def test_lock_timeout_zero_allowed(tmp_path):
    """lock_timeout=0.0 は即時タイムアウトとして有効な値であること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.0)
    # ロックが競合していない場合は RLock は即座に取得できる
    db["key"] = "value"
    assert db["key"] == "value"
    db.close()


# ===========================================================
# P2-3: backup / restore メソッド
# ===========================================================


def test_backup_creates_file(tmp_path):
    """backup() がファイルを作成すること"""
    db_path = str(tmp_path / "src.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["a"] = 1
    db["b"] = {"x": 2}
    db.backup(backup_path)

    assert os.path.exists(backup_path)
    db.close()


def test_backup_content_matches(tmp_path):
    """backup() で作成されたファイルの内容が元 DB と同じであること"""
    db_path = str(tmp_path / "src.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key1"] = "value1"
    db["key2"] = [1, 2, 3]
    db.backup(backup_path)
    db.close()

    # バックアップから読み込んで検証
    db_bak = NanaSQLite(backup_path)
    assert db_bak["key1"] == "value1"
    assert db_bak["key2"] == [1, 2, 3]
    db_bak.close()


def test_restore_replaces_data(tmp_path):
    """restore() 後に DB の内容がバックアップ時点に戻ること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key1"] = "original"
    db.backup(backup_path)

    # バックアップ後にデータを変更
    db["key1"] = "modified"
    db["key2"] = "new_key"
    assert db["key1"] == "modified"

    # リストア
    db.restore(backup_path)
    assert db["key1"] == "original"
    assert "key2" not in db
    db.close()


def test_restore_clears_cache(tmp_path):
    """restore() 後にメモリキャッシュがクリアされること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path, bulk_load=True)
    db["cached_key"] = "cached_value"
    db.backup(backup_path)

    db["new_key"] = "new_value"
    db.restore(backup_path)

    assert not db._all_loaded
    assert "new_key" not in db
    db.close()


def test_restore_on_child_instance_raises(tmp_path):
    """table() で取得した子インスタンスで restore() を呼ぶと NanaSQLiteConnectionError が発生すること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key"] = "value"
    db.backup(backup_path)

    child = db.table("other")
    with pytest.raises(NanaSQLiteConnectionError):
        child.restore(backup_path)

    db.close()


def test_backup_on_closed_db_raises(tmp_path):
    """閉じた DB で backup() を呼ぶと例外が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"
    db.close()

    with pytest.raises(NanaSQLiteClosedError):
        db.backup(str(tmp_path / "backup.db"))


def test_restore_invalid_path_raises(tmp_path):
    """存在しないパスから restore() すると NanaSQLiteDatabaseError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"

    with pytest.raises(NanaSQLiteDatabaseError):
        db.restore(str(tmp_path / "nonexistent.db"))

    db.close()


def test_restore_during_transaction_raises(tmp_path):
    """トランザクション中に restore() を呼ぶと NanaSQLiteTransactionError が発生すること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key"] = "value"
    db.backup(backup_path)

    with pytest.raises(NanaSQLiteTransactionError):
        with db.transaction():
            db["key"] = "in-transaction"
            db.restore(backup_path)  # トランザクション中なので例外が発生するはず

    db.close()


def test_backup_to_memory_path_raises(tmp_path):
    """backup() の dest_path が ':memory:' の場合 NanaSQLiteValidationError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"

    with pytest.raises(NanaSQLiteValidationError):
        db.backup(":memory:")

    db.close()


def test_backup_to_file_memory_uri_raises(tmp_path):
    """backup() の dest_path が 'file::memory:' URI の場合 NanaSQLiteValidationError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"

    with pytest.raises(NanaSQLiteValidationError):
        db.backup("file::memory:?cache=shared")

    db.close()


def test_restore_cleans_stale_wal_sidecar_files(tmp_path):
    """restore() が stale な WAL/SHM/journal サイドカーファイルを削除して正常にリストアすること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key"] = "original"
    db.backup(backup_path)

    # バックアップ後にデータを変更
    db["key"] = "modified"
    db["extra"] = "should_be_gone"

    # Windows では DB 接続が開いている間 SQLite が -shm ファイルをロックするため、
    # サイドカーファイルを作成する前に接続を閉じる必要がある。
    db.close()

    # stale なサイドカーファイルを手動で作成（restore 前に残っている想定）
    stale_marker = b"stale WAL data that should not be replayed"
    for suffix in ("-wal", "-shm", "-journal"):
        sidecar = db_path + suffix
        with open(sidecar, "wb") as f:
            f.write(stale_marker)

    # 再接続してリストアを実行
    db = NanaSQLite(db_path)
    db.restore(backup_path)

    # stale なサイドカー内容が再生されず、バックアップ時点のデータが正しく復元されること
    assert db["key"] == "original"
    assert "extra" not in db

    # -journal は WAL モードでは再生成されないため、削除されたままであること
    assert not os.path.exists(db_path + "-journal")
    # 注: -wal/-shm は WAL モードの再接続で新規作成されるため存在チェックは行わない。
    # 上記のデータ整合性チェック（"original" に戻っていること）で stale WAL が
    # 再生されていないことを間接的に検証している。

    db.close()


def test_restore_on_memory_db_raises(tmp_path):
    """インメモリDB（':memory:'）に対して restore() を呼ぶと NanaSQLiteValidationError が発生すること"""
    # NanaSQLite のコンストラクタが ':memory:' を受け付けるか確認して使用
    backup_path = str(tmp_path / "backup.db")
    # まず通常の DB でバックアップを作成しておく
    src_db = NanaSQLite(str(tmp_path / "src.db"))
    src_db["key"] = "value"
    src_db.backup(backup_path)
    src_db.close()

    # ':memory:' DB に restore() を試みると NanaSQLiteValidationError が発生すること
    mem_db = NanaSQLite(":memory:")
    with pytest.raises(NanaSQLiteValidationError):
        mem_db.restore(backup_path)
    mem_db.close()


# ===========================================================
# 追加テスト: lock_timeout 公開 API 経由での検証
# ===========================================================


def _hold_lock_and_signal(db, lock_held_event, release_event):
    """内部ロックを保持してイベントで通知するヘルパー（スレッド用）"""
    db._lock.acquire()
    try:
        lock_held_event.set()
        release_event.wait()
    finally:
        db._lock.release()


def test_lock_timeout_setitem_raises_lockerror(tmp_path):
    """ロック保持中に __setitem__ を呼ぶと lock_timeout により NanaSQLiteLockError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.2)
    db["existing"] = "value"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            db["new_key"] = "new_value"
    finally:
        release_lock.set()
        t.join(timeout=5.0)
    # キャッシュが更新されていないことを確認（不整合防止）- ロック解放後に確認
    assert "new_key" not in db
    db.close()


def test_lock_timeout_delitem_raises_lockerror(tmp_path):
    """ロック保持中に __delitem__ を呼ぶと lock_timeout により NanaSQLiteLockError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.2)
    db["key_to_delete"] = "value"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            del db["key_to_delete"]
        # キャッシュ・DBが更新されていないことを確認（不整合防止）
        assert "key_to_delete" in db
    finally:
        release_lock.set()
        t.join(timeout=5.0)
    db.close()


def test_lock_timeout_pop_raises_lockerror(tmp_path):
    """ロック保持中に pop() を呼ぶと lock_timeout により NanaSQLiteLockError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.2)
    db["key_to_pop"] = "value"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            db.pop("key_to_pop")
        # キーが残っていることを確認（不整合防止）
        assert "key_to_pop" in db
    finally:
        release_lock.set()
        t.join(timeout=5.0)
    db.close()


def test_lock_timeout_clear_raises_lockerror(tmp_path):
    """ロック保持中に clear() を呼ぶと lock_timeout により NanaSQLiteLockError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.2)
    db["key1"] = "value1"
    db["key2"] = "value2"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            db.clear()
        # データが残っていることを確認（不整合防止）
        assert "key1" in db
        assert "key2" in db
    finally:
        release_lock.set()
        t.join(timeout=5.0)
    db.close()


def test_lock_timeout_restore_raises_lockerror(tmp_path):
    """ロック保持中に restore() を呼ぶと lock_timeout により NanaSQLiteLockError が発生すること"""
    db_path = str(tmp_path / "test.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path, lock_timeout=0.2)
    db["key"] = "original"
    db.backup(backup_path)
    db["key"] = "modified"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            db.restore(backup_path)
        # restoreが失敗してもデータが変わっていないことを確認
        assert db["key"] == "modified"
    finally:
        release_lock.set()
        t.join(timeout=5.0)
    db.close()


def test_lock_timeout_error_message_contains_seconds(tmp_path):
    """NanaSQLiteLockError のメッセージにタイムアウト秒数が含まれること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=1.5)
    errors = []

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        try:
            with db._acquire_lock():
                pass
        except NanaSQLiteLockError as e:
            errors.append(e)
    finally:
        release_lock.set()
        t.join(timeout=5.0)

    assert len(errors) == 1
    assert "1.5s" in str(errors[0])
    db.close()


# ===========================================================
# 追加テスト: backup() 詳細テスト
# ===========================================================


def test_backup_overwrites_existing_file(tmp_path):
    """backup() が既存のバックアップファイルを上書きすること"""
    db_path = str(tmp_path / "src.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key"] = "version1"
    db.backup(backup_path)

    # データを更新して再度バックアップ
    db["key"] = "version2"
    db.backup(backup_path)
    db.close()

    # 最新のバックアップが使われること
    db_bak = NanaSQLite(backup_path)
    assert db_bak["key"] == "version2"
    db_bak.close()


def test_backup_self_copy_raises(tmp_path):
    """backup() の dest_path が現在の DB ファイルと同一の場合 NanaSQLiteValidationError が発生すること"""
    db_path = str(tmp_path / "test.db")
    db = NanaSQLite(db_path)
    db["key"] = "value"

    with pytest.raises(NanaSQLiteValidationError):
        db.backup(db_path)

    db.close()


def test_backup_on_closed_db_raises_closed_error(tmp_path):
    """閉じた DB で backup() を呼ぶと NanaSQLiteClosedError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"
    db.close()

    with pytest.raises(NanaSQLiteClosedError):
        db.backup(str(tmp_path / "backup.db"))


def test_backup_with_multiple_tables(tmp_path):
    """複数テーブルを持つ DB のバックアップが正しく動作すること"""
    db_path = str(tmp_path / "multi.db")
    backup_path = str(tmp_path / "multi_backup.db")

    db = NanaSQLite(db_path)
    db["main_key"] = "main_value"

    users = db.table("users")
    users["alice"] = {"age": 30}
    users["bob"] = {"age": 25}

    db.backup(backup_path)
    db.close()

    # バックアップから両テーブルが読めること
    db_bak = NanaSQLite(backup_path)
    assert db_bak["main_key"] == "main_value"

    users_bak = db_bak.table("users")
    assert users_bak["alice"] == {"age": 30}
    assert users_bak["bob"] == {"age": 25}
    db_bak.close()


def test_backup_non_blocking_allows_concurrent_write(tmp_path):
    """backup() はノンブロッキングであり、実行中に他のスレッドが NanaSQLite に書き込めること"""
    db_path = str(tmp_path / "test.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path, lock_timeout=5.0)
    for i in range(100):
        db[f"key{i}"] = i

    write_errors = []
    write_success = []

    def concurrent_write():
        """backup() 中にも書き込みできることを確認するスレッド"""
        try:
            db["concurrent_write"] = "written_during_backup"
            write_success.append(True)
        except Exception as e:
            write_errors.append(e)

    backup_started = threading.Event()
    original_backup = db.backup

    def backup_with_signal(dest):
        backup_started.set()
        original_backup(dest)

    db.backup = backup_with_signal

    backup_thread = threading.Thread(target=db.backup, args=(backup_path,))
    backup_thread.start()
    backup_started.wait()

    write_thread = threading.Thread(target=concurrent_write)
    write_thread.start()
    write_thread.join(timeout=5.0)

    backup_thread.join(timeout=10.0)

    assert not write_errors, f"Concurrent write failed: {write_errors}"
    assert write_success, "Concurrent write did not succeed"

    # Restore original method
    db.backup = original_backup
    db.close()


def test_backup_large_data(tmp_path):
    """大量データを持つ DB の backup() が正しく完了すること"""
    db_path = str(tmp_path / "large.db")
    backup_path = str(tmp_path / "large_backup.db")

    db = NanaSQLite(db_path)
    # 500件のデータを書き込む
    for i in range(500):
        db[f"key_{i:04d}"] = {"index": i, "data": f"value_{i}" * 10}

    db.backup(backup_path)
    db.close()

    assert os.path.exists(backup_path)
    db_bak = NanaSQLite(backup_path)
    assert db_bak["key_0000"] == {"index": 0, "data": "value_0" * 10}
    assert db_bak["key_0499"] == {"index": 499, "data": "value_499" * 10}
    assert len(db_bak) == 500
    db_bak.close()


# ===========================================================
# 追加テスト: restore() 詳細テスト
# ===========================================================


def test_restore_updates_child_table_connection(tmp_path):
    """restore() 後に table() で取得した子インスタンスも新しい接続を使うこと"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key"] = "original"
    users = db.table("users")
    users["alice"] = "user1"

    db.backup(backup_path)

    # バックアップ後に変更
    db["key"] = "modified"
    users["bob"] = "user2"

    # リストア
    db.restore(backup_path)

    # 親インスタンスのデータが戻ること
    assert db["key"] == "original"
    # 子インスタンスも更新されること
    assert "bob" not in users
    assert users["alice"] == "user1"
    db.close()


def test_restore_on_closed_db_raises(tmp_path):
    """閉じた DB で restore() を呼ぶと NanaSQLiteClosedError が発生すること"""
    db_path = str(tmp_path / "test.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key"] = "value"
    db.backup(backup_path)
    db.close()

    with pytest.raises(NanaSQLiteClosedError):
        db.restore(backup_path)


def test_restore_allows_subsequent_operations(tmp_path):
    """restore() 後も通常のCRUD操作が正しく動作すること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["key1"] = "value1"
    db.backup(backup_path)

    db["key1"] = "changed"
    db["key2"] = "extra"
    db.restore(backup_path)

    # リストア後に CRUD が正常に動くこと
    assert db["key1"] == "value1"
    assert "key2" not in db
    db["key3"] = "new_after_restore"
    assert db["key3"] == "new_after_restore"
    del db["key1"]
    assert "key1" not in db
    db.close()


def test_restore_with_bulk_load_mode(tmp_path):
    """bulk_load モードで restore() が正しくキャッシュをクリアすること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path, bulk_load=True)
    db["a"] = 1
    db["b"] = 2
    db.load_all()  # キャッシュに全データをロード
    assert db._all_loaded

    db.backup(backup_path)

    db["c"] = 3
    db.restore(backup_path)

    # キャッシュがクリアされること
    assert not db._all_loaded
    assert "c" not in db
    assert db["a"] == 1
    db.close()


def test_restore_preserves_table_name(tmp_path):
    """restore() 後もテーブル名設定が保持されること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path, table="my_table")
    db["key"] = "value"
    db.backup(backup_path)

    db["key"] = "modified"
    db.restore(backup_path)

    assert db["key"] == "value"
    assert db._table == "my_table"
    db.close()


def test_restore_file_memory_uri_raises(tmp_path):
    """file::memory: URI の DB パスは _is_in_memory_path() が True を返し restore() が拒否すること"""
    backup_path = str(tmp_path / "backup.db")
    src_db = NanaSQLite(str(tmp_path / "src.db"))
    src_db["key"] = "value"
    src_db.backup(backup_path)
    src_db.close()

    # _is_in_memory_path() が "file::memory:..." を True と判定することを確認する
    # (実際のインメモリDB接続は ":memory:" テストでカバー済みのため、
    #  ここでは内部判定メソッドのみを検証する)
    from nanasqlite.core import NanaSQLite as _NanaSQLite
    assert _NanaSQLite._is_in_memory_path("file::memory:?cache=shared")
    assert _NanaSQLite._is_in_memory_path("file::memory:")
    assert not _NanaSQLite._is_in_memory_path(str(tmp_path / "real.db"))


# ===========================================================
# 追加テスト: キャッシュ・DB 整合性テスト
# ===========================================================


def test_cache_consistency_after_setitem_lock_timeout(tmp_path):
    """lock_timeout で __setitem__ が失敗した後もキャッシュと DB の整合性が保たれること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.1)
    db["existing"] = "original_value"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            db["existing"] = "should_not_update"
    finally:
        release_lock.set()
        t.join(timeout=5.0)

    # キャッシュと DB が元の値のまま
    assert db["existing"] == "original_value"
    # 再接続して確認
    db.close()
    db2 = NanaSQLite(str(tmp_path / "test.db"))
    assert db2["existing"] == "original_value"
    db2.close()


def test_cache_consistency_after_delitem_lock_timeout(tmp_path):
    """lock_timeout で __delitem__ が失敗した後もキャッシュと DB の整合性が保たれること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.1)
    db["key_to_keep"] = "important"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            del db["key_to_keep"]
    finally:
        release_lock.set()
        t.join(timeout=5.0)

    # キーが残っていること
    assert "key_to_keep" in db
    assert db["key_to_keep"] == "important"
    db.close()
    db2 = NanaSQLite(str(tmp_path / "test.db"))
    assert db2["key_to_keep"] == "important"
    db2.close()


def test_sequential_operations_after_lock_timeout_recovery(tmp_path):
    """lock_timeout 後のリカバリ: ロック解放後は通常操作が再開できること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.1)
    db["key"] = "initial"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()

    # ロック保持中は失敗する
    with pytest.raises(NanaSQLiteLockError):
        db["key"] = "during_lock"

    # ロック解放
    release_lock.set()
    t.join(timeout=5.0)

    # ロック解放後は通常通り動作すること
    db["key"] = "after_lock"
    assert db["key"] == "after_lock"
    db.close()


# ===========================================================
# 追加テスト: lock_timeout のエッジケース
# ===========================================================


def test_lock_timeout_integer_accepted(tmp_path):
    """lock_timeout に整数値を渡した場合も正しく動作すること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=2)
    db["key"] = "value"
    assert db["key"] == "value"
    db.close()


def test_lock_timeout_very_small_value(tmp_path):
    """lock_timeout に非常に小さい値（0.001秒）を設定してもエラーにならないこと"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=0.001)
    db["key"] = "value"
    assert db["key"] == "value"
    db.close()


def test_lock_timeout_is_stored_on_instance(tmp_path):
    """lock_timeout の値がインスタンスに正しく格納されること"""
    db = NanaSQLite(str(tmp_path / "test.db"), lock_timeout=3.5)
    assert db._lock_timeout == 3.5
    db.close()


def test_lock_timeout_none_stored_on_instance(tmp_path):
    """lock_timeout=None（デフォルト）がインスタンスに正しく格納されること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    assert db._lock_timeout is None
    db.close()


# ===========================================================
# 追加テスト: backup()/restore() の NanaSQLiteLockError 検証
# （lock_timeout 設定時のロック取得タイムアウト）
# ===========================================================


def test_backup_raises_lockerror_when_lock_held(tmp_path):
    """ロック保持中に backup() を呼ぶと lock_timeout により NanaSQLiteLockError が発生すること"""
    db_path = str(tmp_path / "test.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path, lock_timeout=0.2)
    db["key"] = "value"

    lock_held = threading.Event()
    release_lock = threading.Event()
    t = threading.Thread(target=_hold_lock_and_signal, args=(db, lock_held, release_lock))
    t.start()
    lock_held.wait()
    try:
        with pytest.raises(NanaSQLiteLockError):
            db.backup(backup_path)
        # バックアップファイルが作成されていないこと
        assert not os.path.exists(backup_path)
    finally:
        release_lock.set()
        t.join(timeout=5.0)
    db.close()


# ===========================================================
# 追加テスト: __delitem__ / clear() の _check_connection() 検証
# （クローズ済み接続での NanaSQLiteClosedError 送出）
# ===========================================================


def test_delitem_on_closed_db_raises_closed_error(tmp_path):
    """`del db[key]` をクローズ済み接続で呼ぶと NanaSQLiteClosedError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key"] = "value"
    db.close()

    with pytest.raises(NanaSQLiteClosedError):
        del db["key"]


def test_clear_on_closed_db_raises_closed_error(tmp_path):
    """`db.clear()` をクローズ済み接続で呼ぶと NanaSQLiteClosedError が発生すること"""
    db = NanaSQLite(str(tmp_path / "test.db"))
    db["key1"] = "value1"
    db["key2"] = "value2"
    db.close()

    with pytest.raises(NanaSQLiteClosedError):
        db.clear()


# ===========================================================
# 追加テスト: backup() 非ブロッキング動作の検証
# ===========================================================


def test_backup_does_not_block_same_db_writes(tmp_path):
    """backup() は NanaSQLite 内部ロックを長時間保持しないため、
    バックアップ完了前に同一インスタンスへの書き込みが可能であること"""
    db_path = str(tmp_path / "test.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    db["initial"] = "data"

    # バックアップを実行してからも書き込みが可能であること
    db.backup(backup_path)
    db["after_backup"] = "still_works"
    assert db["after_backup"] == "still_works"
    db.close()


def test_restore_data_integrity_after_multiple_backups(tmp_path):
    """複数回バックアップして restore() すると最後にリストアしたバックアップが反映されること"""
    db_path = str(tmp_path / "main.db")
    backup1_path = str(tmp_path / "backup1.db")
    backup2_path = str(tmp_path / "backup2.db")

    db = NanaSQLite(db_path)
    db["key"] = "v1"
    db.backup(backup1_path)

    db["key"] = "v2"
    db.backup(backup2_path)

    db["key"] = "v3"

    # backup1 でリストア
    db.restore(backup1_path)
    assert db["key"] == "v1"

    # backup2 でリストア
    db.restore(backup2_path)
    assert db["key"] == "v2"

    db.close()


def test_backup_restore_roundtrip_with_complex_data(tmp_path):
    """複雑なデータ（リスト、ネスト辞書、None等）のバックアップ/リストアが正確に動作すること"""
    db_path = str(tmp_path / "main.db")
    backup_path = str(tmp_path / "backup.db")

    db = NanaSQLite(db_path)
    complex_data = {
        "list": [1, 2, 3, {"nested": True}],
        "dict": {"a": 1, "b": {"c": 2}},
        "none_val": None,
        "bool_val": True,
        "int_val": 42,
        "float_val": 3.14,
    }
    db["complex"] = complex_data
    db["simple"] = "string"

    db.backup(backup_path)

    db["complex"] = "changed"
    db["extra"] = "added"

    db.restore(backup_path)

    assert db["complex"] == complex_data
    assert db["simple"] == "string"
    assert "extra" not in db
    db.close()



"""Documentation regression tests for the validation guide."""

from pathlib import Path

REPO_ROOT = Path(__file__).parent.parent


def assert_contains_all(path: Path, snippets: list[str]) -> None:
    """Assert that a documentation file contains all expected snippets."""
    content = path.read_text(encoding="utf-8")
    for snippet in snippets:
        assert snippet in content, f"Expected to find {snippet!r} in {path}"


class TestStructuredDocsValidationGuide:
    """The parallel docs/en and docs/ja trees should expose the validation guide."""

    def test_en_validation_guide_exists_with_core_topics(self) -> None:
        guide = REPO_ROOT / "docs" / "en" / "guide" / "validation.md"
        assert guide.exists(), f"Missing validation guide: {guide}"
        assert_contains_all(
            guide,
            [
                "validkit-py",
                "pip install nanasqlite[validation]",
                "coerce=True",
                "batch_update()",
                "HAS_VALIDKIT",
            ],
        )

    def test_ja_validation_guide_exists_with_core_topics(self) -> None:
        guide = REPO_ROOT / "docs" / "ja" / "guide" / "validation.md"
        assert guide.exists(), f"Missing validation guide: {guide}"
        assert_contains_all(
            guide,
            [
                "validkit-py",
                "pip install nanasqlite[validation]",
                "coerce=True",
                "batch_update()",
                "HAS_VALIDKIT",
            ],
        )

    def test_readmes_link_to_validation_guide(self) -> None:
        assert_contains_all(
            REPO_ROOT / "docs" / "en" / "README.md",
            ["guide/validation.md", "[Validation]"],
        )
        assert_contains_all(
            REPO_ROOT / "docs" / "ja" / "README.md",
            ["guide/validation.md", "[バリデーション]"],
        )


class TestSiteValidationGuide:
    """The published VitePress site should expose the validation guide."""

    def test_site_validation_guides_exist(self) -> None:
        assert_contains_all(
            REPO_ROOT / "docs" / "site" / "en" / "validation_guide.md",
            ["validkit-py", "pip install nanasqlite[validation]", "coerce=True"],
        )
        assert_contains_all(
            REPO_ROOT / "docs" / "site" / "validation_guide.md",
            ["validkit-py", "pip install nanasqlite[validation]", "coerce=True"],
        )

    def test_site_sidebar_links_include_validation_guide(self) -> None:
        config = REPO_ROOT / "docs" / "site" / ".vitepress" / "config.mts"
        assert_contains_all(
            config,
            [
                "/validation_guide",
                "/en/validation_guide",
            ],
        )



import pytest

from nanasqlite import AsyncNanaSQLite, NanaSQLiteValidationError


@pytest.mark.asyncio
async def test_query_reserved_keyword_column(tmp_path):
    """
    AsyncNanaSQLite.query で予約語（group）をカラム名に使用できることを検証
    識別子サニタイズにより、予約語が引用符で適切にエスケープされることを確認
    """
    db_path = str(tmp_path / "test_sanitization.db")
    db = AsyncNanaSQLite(db_path)
    await db.create_table("test", {"group": "TEXT", "name": "TEXT"})
    await db.sql_insert("test", {"group": "Admin", "name": "Alice"})

    # 予約語カラムを含むクエリが成功することを検証
    results = await db.query(table_name="test", columns=["group", "name"])
    assert len(results) == 1
    assert results[0]["group"] == "Admin"
    assert results[0]["name"] == "Alice"

    # 識別子が適切に引用符でエスケープされていることを検証
    # _sanitize_identifierの動作を間接的に確認
    from nanasqlite import NanaSQLite

    assert NanaSQLite._sanitize_identifier("group") == '"group"'
    assert NanaSQLite._sanitize_identifier("name") == '"name"'

    await db.close()


@pytest.mark.asyncio
async def test_query_per_query_validation(tmp_path):
    """
    AsyncNanaSQLite.query でクエリ単位の関数許可（allowed_sql_functions）が機能することを検証
    """
    db_path = str(tmp_path / "test_validation.db")
    db = AsyncNanaSQLite(db_path, strict_sql_validation=True)
    await db.create_table("test", {"name": "TEXT"})
    await db.sql_insert("test", {"name": "Alice"})

    # HEX はデフォルトで許可されていないためエラーになるはず
    with pytest.raises((NanaSQLiteValidationError, ValueError)):
        await db.query(table_name="test", columns=["HEX(name)"])

    # クエリ単位で許可すれば成功するはず
    results = await db.query(table_name="test", columns=["HEX(name)"], allowed_sql_functions=["HEX"])
    hex_value = results[0]["HEX(name)"]
    assert hex_value == "416C696365"
    # HEX値をデコードして元の文字列と一致することを確認 (#4)
    assert bytes.fromhex(hex_value).decode("utf-8") == "Alice"
    await db.close()


@pytest.mark.asyncio
async def test_groupby_validation_in_pagination(tmp_path):
    """
    query_with_pagination で group_by 句のバリデーションが機能することを検証
    """
    db_path = str(tmp_path / "test_groupby.db")
    db = AsyncNanaSQLite(db_path, strict_sql_validation=True)
    await db.create_table("test", {"category": "TEXT", "val": "INTEGER"})
    await db.sql_insert("test", {"category": "A", "val": 10})

    # HEX を group_by に含めるとエラーになるはず
    with pytest.raises((NanaSQLiteValidationError, ValueError)):
        await db.query_with_pagination(table_name="test", columns=["category", "SUM(val)"], group_by="HEX(category)")

    # 許可すれば成功する
    results = await db.query_with_pagination(
        table_name="test", columns=["category", "SUM(val)"], group_by="HEX(category)", allowed_sql_functions=["HEX"]
    )
    assert len(results) == 1
    await db.close()


@pytest.mark.asyncio
async def test_complex_column_aliases_metadata(tmp_path):
    """
    cursor.description を使用したカラム名抽出が複雑な式でも正常に動作することを検証 (#2)
    """
    db_path = str(tmp_path / "test_complex_cols.db")
    db = AsyncNanaSQLite(db_path)
    await db.create_table("test", {"name": "TEXT", "val": "INTEGER"})
    await db.sql_insert("test", {"name": "Alice as Bob", "val": 10})

    # " as " を含む文字列リテラルや複雑なエイリアス
    # 以前の正規表現パースでは "Alice as Bob" の " as " で誤分割される可能性があった
    sql_cols = ['name as "user_name"', "val + 100 as total", "'prefix as ' || name as complex_label"]
    results = await db.query(table_name="test", columns=sql_cols)

    assert len(results) == 1
    assert results[0]["user_name"] == "Alice as Bob"
    assert results[0]["total"] == 110
    assert results[0]["complex_label"] == "prefix as Alice as Bob"
    await db.close()


def test_eq_when_closed(tmp_path):
    """
    DB接続が閉じている場合に == 演算子が NanaSQLiteClosedError を発生させることを検証 (#5)
    """
    from nanasqlite import NanaSQLite, NanaSQLiteClosedError

    db_path = str(tmp_path / "test_eq_closed.db")
    db1 = NanaSQLite(db_path)
    db1["k"] = "v"

    db2 = {"k": "v"}

    def _compare_db1_db2():
        return db1 == db2

    assert _compare_db1_db2() is True  # 接続中は True

    db1.close()
    # 閉じられた接続での等価性チェックは NanaSQLiteClosedError を発生させる
    # これは他のマジックメソッド（__getitem__, __setitem__, __delitem__）と一貫性があります
    with pytest.raises(NanaSQLiteClosedError):
        _ = _compare_db1_db2()



import pytest

from nanasqlite import AsyncNanaSQLite


@pytest.fixture
def db_path(tmp_path):
    return str(tmp_path / "test_aliases.db")


@pytest.mark.asyncio
async def test_async_get_alias(db_path):
    """get エイリアスが aget と同じ動作をすることを確認"""
    async with AsyncNanaSQLite(db_path) as db:
        await db.aset("test_key", {"value": "test_data"})

        # aget を使用
        result_aget = await db.aget("test_key")
        # get エイリアスを使用
        result_get = await db.get("test_key")

        assert result_aget == result_get
        assert result_get == {"value": "test_data"}


@pytest.mark.asyncio
async def test_async_contains_alias(db_path):
    """contains エイリアスが acontains と同じ動作をすることを確認"""
    async with AsyncNanaSQLite(db_path) as db:
        await db.aset("existing_key", "data")

        # acontains を使用
        result_acontains = await db.acontains("existing_key")
        # contains エイリアスを使用
        result_contains = await db.contains("existing_key")

        assert result_acontains == result_contains
        assert result_contains is True

        # 存在しないキー
        assert await db.contains("nonexistent") is False


@pytest.mark.asyncio
async def test_async_keys_alias(db_path):
    """keys エイリアスが akeys と同じ動作をすることを確認"""
    async with AsyncNanaSQLite(db_path) as db:
        await db.aset("key1", "value1")
        await db.aset("key2", "value2")
        await db.aset("key3", "value3")

        # akeys を使用
        result_akeys = await db.akeys()
        # keys エイリアスを使用
        result_keys = await db.keys()

        assert set(result_akeys) == set(result_keys)
        assert set(result_keys) == {"key1", "key2", "key3"}


@pytest.mark.asyncio
async def test_async_values_alias(db_path):
    """values エイリアスが avalues と同じ動作をすることを確認"""
    async with AsyncNanaSQLite(db_path) as db:
        await db.aset("key1", "value1")
        await db.aset("key2", "value2")

        # avalues を使用
        result_avalues = await db.avalues()
        # values エイリアスを使用
        result_values = await db.values()

        assert result_avalues == result_values
        assert set(result_values) == {"value1", "value2"}


@pytest.mark.asyncio
async def test_async_items_alias(db_path):
    """items エイリアスが aitems と同じ動作をすることを確認"""
    async with AsyncNanaSQLite(db_path) as db:
        await db.aset("key1", "value1")
        await db.aset("key2", "value2")

        # aitems を使用
        result_aitems = await db.aitems()
        # items エイリアスを使用
        result_items = await db.items()

        assert result_aitems == result_items
        assert set(result_items) == {("key1", "value1"), ("key2", "value2")}
